# BELONG TO SERVICE TASK
import json, uuid, time, pytz, os, logging, re
from datetime import datetime
from enum import Enum
from io import StringIO
from psycopg2.extras import RealDictCursor
from enums import SupportedOS, DBType, ClusterRoles, RestoreOptions, Status, Severity, Inventory, Schedule, LicenseType, Timeout
from tasks_core import get_postgres_connection, get_date_modified_folder, logger, remote_command, create_winrm_connection, remote_command_win,\
    connect_to_sqlserver_instance, generate_complex_password, create_ssh_connection, TASK_TMP_FILES_DIR_WIN

class BackupType(Enum):
    FULL = 'full'
    LOG = 'log'
    COPY_ONLY = 'copy_only'
    DIFFERENTIAL = 'differential'
    SNAPSHOT = 'snapshot'

class BackupEncryption(Enum): # best down to worse
    AES_256 = 'aes_256'
    AES_192 = 'aes_192'
    AES_128 = 'aes_128'
    NO_ENCRYPTION = 'no_encryption'

class RecoveryModel(Enum):
    FULL = 'full'
    BULK_LOGGED = 'bulk_logged'
    SIMPLE = 'simple'

class DatabaseState(Enum):
    ONLINE = 'online'
    OFFLINE = 'offline'
    RESTORING = 'restoring'
    EMERGENCY = 'emergency'

class UserAccess(Enum):
    SINGLE_USER = 'single_user'
    RESTRICTED_USER = 'restricted_user'
    MULTI_USER = 'multi_user'

class ValidationResult(Enum):
    VALID = 'valid'
    INVALID = 'invalid'
    ERROR = 'error'

AZURITE_SERVICE = 'azurite.service'
SYSTEM_DATABASES = ['master', 'msdb', 'model', 'tempdb']
BACKUP_CERTIFICATE = {
    "name": '123cluster_certificate',
    "subject": '123cluster database backup certificate',
    "password": 'LAl&P?D9daR!ijCtye@W'
}

def prepare_encryptors(cur, certificate_name, certificate_subject, password):
    """
    Helper procedure of encrypted operations, exceptions are handled in parent functions
    """
    
    """
    Encryptor are either certificate (preferred, because exportable) OR asymmetric key.
    Not both.
    Self-signed certificates (ENCRYPTION BY PASSWORD) could not be used for backup,
    backup certificates should be signed by cryptographic key.
    Preffered is database master key,because exportable and dont need key file
    """
    # database master key name are created by system and always same
    cur.execute("SELECT name FROM sys.symmetric_keys WHERE name = N'##MS_DatabaseMasterKey##'")
    res = cur.fetchall()
    master_key_exists = True if res and res[0] else False
    # certificate names ARE case-sensitive
    cur.execute(f"SELECT name FROM sys.certificates WHERE name = N'{certificate_name}'")
    res = cur.fetchall()
    certificate_exists = True if res and res[0] else False
    if not master_key_exists:
        cur.execute(f"CREATE MASTER KEY ENCRYPTION BY PASSWORD = N'{password}'")
    
    if not certificate_exists:
        cur.execute(f"""CREATE CERTIFICATE [{certificate_name}] WITH SUBJECT = N'{certificate_subject}'""")

def wait_busy_cursor(cur):
    "Helper procedure for database backup/restore, exceptions are handled in parent functions"
    """wait with pyodbc cursor while timely operation are in progress"""
    if cur:
        while cur.nextset():
            pass

def drop_database(host, node, db_name, recreate = False):
     """
     Drop sqlserver database.
     recreate = True - recreate database after drop.
     """
     try:
        global SYSTEM_DATABASES
        db = db_name.strip()
        if db in SYSTEM_DATABASES:
            # return error - since this can happen by accident,or user malice/lack of knowledge
            return (True, f'{db} are system database.Dropping it may lead to corruption of instance.')
        
        conn = connect_to_sqlserver_instance(host['name'], 'master', node['database_user'], node['database_pwd'], node['port'])
        if isinstance(conn, tuple):
            return (True, str(conn[1]))
        
        with conn:
            conn.autocommit = True
            with conn.cursor() as cur:
                cur.execute(f"IF EXISTS (SELECT database_id FROM sys.databases WHERE name = N'{db}') BEGIN ALTER DATABASE [{db}] SET single_user WITH ROLLBACK IMMEDIATE;ALTER DATABASE [{db}] SET OFFLINE WITH ROLLBACK IMMEDIATE;DROP DATABASE [{db}] END")
                if recreate:
                    is_error, result = ensure_database_exists(host, node, db_name)
                    if is_error:
                        return is_error, result
                
                cur.execute("SELECT SUM(mf.size * 8) AS SizeKB FROM sys.master_files mf JOIN sys.databases d ON mf.database_id = d.database_id WHERE mf.database_id > 4 GROUP BY mf.database_id")
                res = cur.fetchone()
                node_size = res[0] if res and res[0] else 0.0
                
                return (False, node_size)
     except Exception as error:
         return (True, str(error))

def ensure_database_exists(host, node, db_name, recovery_model=RecoveryModel.FULL.value):
     """
     Ensure that database exists on instance.
     Helper function for database creation.
     recovery_mode - whether to use a specific recovery mode for database
     """
     try:
        global SYSTEM_DATABASES
        db = db_name.strip()
        # system databases exists for sure
        if db in SYSTEM_DATABASES:
            return (False, '')
        
        conn = connect_to_sqlserver_instance(host['name'], 'master', node['database_user'], node['database_pwd'], node['port'])
        if isinstance(conn, tuple):
            return (True, str(conn[1]))
        
        with conn:
            conn.autocommit = True
            with conn.cursor() as cur:
                cur.execute(f"IF NOT EXISTS (SELECT database_id FROM sys.databases WHERE name = N'{db}') BEGIN CREATE DATABASE [{db}];ALTER DATABASE [{db}] SET multi_user;ALTER DATABASE [{db}] SET RECOVERY {recovery_model} END")
                return (False, '')
     except Exception as error:
        return (True, str(error))

def share_database_access(host, node, db_name):
     """
     Helper function for export/extract.
     Allow database to be accessed by several sessions at once.
     """
     try:
        db = db_name.strip()
        
        conn = connect_to_sqlserver_instance(host['name'], 'master', node['database_user'], node['database_pwd'], node['port'])
        if isinstance(conn, tuple):
            return (True, str(conn[1]))
        
        with conn:
            conn.autocommit = True
            with conn.cursor() as cur:
                cur.execute(f"IF EXISTS (SELECT * FROM sys.databases WHERE name = '{db}') BEGIN ALTER DATABASE [{db}] SET multi_user END")
                return (False, '')
     except Exception as error:
        return (True, str(error))

def execute_powershell_command(linux_host, command):
    """execute powershell command from linux"""

    try:
        client = create_ssh_connection(linux_host['name'], linux_host['username'], linux_host['password'], linux_host['private_key'], int(linux_host['ssh_port']))
        if isinstance(client, tuple):
            return ('', str(client[1]), 1)
        
        with client:
            with StringIO(command) as file_obj_cmd:
                # Run the Python file-like object using exec_command
                stdin, stdout, stderr = client.exec_command("pwsh")
                if 'not found' in str(stdout):
                    return ('', f"pwsh command has not found on {linux_host['name']}", 1)

                stdin.write(file_obj_cmd.read())
                stdin.channel.shutdown_write()
                exit_status = stdout.channel.recv_exit_status()# Blocking call
                try:
                    stderr = stderr.read().decode('ascii').strip()
                    stdout = stdout.read().decode('ascii').split('\n')
                except UnicodeDecodeError:
                    stderr = stderr.read().decode('utf-8').strip()
                    stdout = stdout.read().decode('utf-8').split('\n')
                
                return stdout,stderr,exit_status
    except Exception as error:
        return ('', f"execute_powershell_command error: {str(error)}", 1)

def copy_files_linux_win(linux_host, win_host, linux_path, win_path, to_linux=True):
    # """Prerequisites on linux:
    #     https://learn.microsoft.com/en-us/powershell/scripting/install/installing-powershell-on-linux?view=powershell-7.4
    #     install-Module -name PSWSMan (or install-wsman)
    #     yum install gssntlmssp
    #     (or apt-get install gss-ntlmssp unixodbc curl
    #         /path/to/odbc_install.sh)
    #    Prerequisites on win:
    #     whoami => domain\username => username@domain
    #     $WinUsername = "$env:USERNAME@$env:COMPUTERNAME"
    #     Set-Item WSMan:\localhost\Client\TrustedHosts -Value "Linux IP" -Force
    #     enable-psremoting -force
    # """
    # to_linux means from win to linux else from linux to win
    try:
        domain = win_host['ex_params']['domain']
        commands = f'''$password = ConvertTo-SecureString "{win_host['password']}" -AsPlainText -Force
$cred = New-Object System.Management.Automation.PSCredential ("{win_host['username']}@{domain}", $password)
$session = New-PSSession -ComputerName "{win_host['name']}" -Port {win_host['ssh_port']} -Credential $cred -Authentication Negotiate\n'''
        commands += f'Copy-Item -Path "{win_path}" -Destination "{linux_path}" -FromSession $session -Recurse -Force' if to_linux else f'Copy-Item "{linux_path}" -Destination "{win_path}" -ToSession $session -Recurse -Force'
        stdout, stderr, exit_status = execute_powershell_command(linux_host, commands)

        if stderr:
            return (True, str(stderr))
    except Exception as error:
        return (True, f"copy_files_linux_win error: {str(error)}")
    
    return (False, str(stdout))

def find_sql_package(host) -> tuple:
    """find location of sqlpackage on Linux"""
    # https://learn.microsoft.com/en-us/sql/tools/sqlpackage/sqlpackage-download?view=sql-server-ver16
    # ln -s /usr/local/sqlpackage /usr/local/bin/
    
    sqlpackage_bin = '/usr/local/bin/sqlpackage'
    try:
        jsn = json.loads(remote_command(host, f'sudo which sqlpackage', True), strict=False)
        if jsn['stdout'] != '':
            sqlpackage_bin = jsn['stdout'].strip()
        
        if jsn['stderr'] != '' and 'permission denied' not in jsn['stderr'].lower():
            logging.error(f"find_sql_package error: {host['name']}: {jsn['stderr']}")
            return sqlpackage_bin
    except Exception as error:
        logging.error(f"find_sql_package error: {host['name']}: {str(error)}")
        return sqlpackage_bin

    return sqlpackage_bin

# def find_sql_package(linux_host, win_host):
# https://learn.microsoft.com/en-us/sql/tools/sqlpackage/sqlpackage-download?view=sql-server-ver16
#     try:
#         client = create_ssh_connection(linux_host['name'], linux_host['username'], linux_host['password'], linux_host['private_key'], int(linux_host['ssh_port']))
#         if isinstance(client, tuple):
#             return (True, str(client[1]))

#         commands = f'''$password = ConvertTo-SecureString "{win_host['password']}" -AsPlainText -Force
# $Cred = New-Object System.Management.Automation.PSCredential ("{win_host['username']}", $password)
# $Session = New-PSSession -ComputerName "{win_host['name']}" -Credential $Cred -Authentication Negotiate\n'''
#         # Run PowerShell command to find SqlPackage.exe
#         commands += 'Get-Command -Name SqlPackage.exe -ErrorAction Stop'

#         stdout, stderr, exit_status = execute_powershell_command(linux_host, commands)

#         if stderr:
#             return (True, str(stderr))
#     except Exception as error:
#         return (True, f"find_sql_package error: {str(error)}")
    
#     return (False, str(stdout))

def create_new_subscription(db, subscriber_sql_name, subscriber_db_name, subscriber_user, subscriber_pwd, publication_name, publication_type = "transactional", publication_operation = "Push") -> str:

    if publication_name == None:
        publication_name = f"{db['name']}_pub"

    if publication_type == "merge":
        subscription_procedure = "sp_addmergesubscription"
        agent_procedure = "sp_addmergesubscription_agent"
    elif publication_operation == "pull":
        subscription_procedure = "sp_addpullsubscription"
        agent_procedure = "sp_addpullsubscription_agent"
    else:
        subscription_procedure = "sp_addsubscription"
        agent_procedure = "sp_addpushsubscription_agent"

    return f'''
    USE [{db["name"]}]
    EXEC {subscription_procedure}
    @publication = N'{publication_name}',
    @subscriber = N'{subscriber_sql_name}',
    @destination_db = N'{subscriber_db_name}',
    @subscription_type = N'{publication_operation}',
    @sync_type = N'automatic',
    @article = N'all',
    @update_mode = N'read only',
    @subscriber_type = 0
    EXEC {agent_procedure}
    @publication = N'{publication_name}',
    @subscriber = N'{subscriber_sql_name}',
    @subscriber_db = N'{subscriber_db_name}',
    @subscriber_security_mode = 0,
    @subscriber_login = N'{subscriber_user}',
    @subscriber_password = N'{subscriber_pwd}',
    @frequency_type = 1,
    @frequency_interval = 0,
    @frequency_relative_interval = 0,
    @frequency_recurrence_factor = 0,
    @frequency_subday = 0,
    @frequency_subday_interval = 0,
    @active_start_time_of_day = 0,
    @active_end_time_of_day = 0,
    @active_start_date = 0,
    @active_end_date = 19950101
    '''

def backup_sqlserver_database(host, node, db_name, backup_folder, backup_name, backup_type = BackupType.FULL.value, encryption = BackupEncryption.AES_128.name, password = '') -> str:
  
    try:
        conn = connect_to_sqlserver_instance(host['name'], 'master', node['database_user'], node['database_pwd'], node['port'])
        if isinstance(conn, tuple):
            return str(conn[1])
        
        with conn:
            conn.autocommit = True
            with conn.cursor() as cur:
                # check if passed backup mode are compatible with db recovery model
                db = db_name.strip()
                cur.execute(f"SELECT recovery_model_desc, user_access_desc FROM sys.databases WHERE name = N'{db}'")
                res = cur.fetchone()
                recovery_model = RecoveryModel.SIMPLE.value if not res else res[0].strip().lower()
                user_access = ValidationResult.ERROR.value if not res else res[1].strip().lower()
                global BACKUP_CERTIFICATE
                backup_type_lower = backup_type.strip().lower()
                name = backup_name.strip()
                # must be always uppercase
                encryption_upper = encryption.strip().upper()
                if RecoveryModel.SIMPLE.value not in recovery_model and BackupType.LOG.value in backup_type_lower:
                    what = "LOG"
                    extension = ".trn"
                else:
                    what = "DATABASE"
                    extension = ".bak"
                
                formatted_name = f'{name}@{BackupType.FULL.value}' if RecoveryModel.SIMPLE.value in recovery_model and BackupType.LOG.value in backup_type_lower else f'{name}@{backup_type_lower}'
                filename = f"{formatted_name}{extension}"
                folder = backup_folder.strip()
                backup_path = f'{folder}\\{filename}'
                backup_date = datetime.utcnow()
                backup_command = f'''
                    BACKUP {what} [{db}]
                    TO DISK = N'{backup_path}'
                    WITH FORMAT, INIT,
                    name = N'{what.lower()} backup of database {db}',
                    medianame = N'{db}_{backup_date}',
                    description = N'Backup of {db},started at {backup_date}'
                    '''
                
                if BackupType.DIFFERENTIAL.value in backup_type_lower and RecoveryModel.SIMPLE.value not in recovery_model:
                    backup_command += f",\r\n DIFFERENTIAL"
                
                if BackupType.COPY_ONLY.value in backup_type_lower:
                    backup_command += f",\r\n COPY_ONLY"
                
                should_encrypt = isinstance(password, str) and password != '' and BackupEncryption.NO_ENCRYPTION.name not in encryption_upper
                if should_encrypt:
                    """
                    ENCRYPTION options:
                    - AES_128
                    - AES_192 - x
                    - AES_256
                    - NO_ENCRYPTION - could not be set as encryption algorithm,
                    just means that backup is not encrypted.
                    Other options are deprecated, can be used only in compatibility mode 120.
                    """
                    backup_command += f",\r\n ENCRYPTION (ALGORITHM = {encryption_upper}, \r\n SERVER CERTIFICATE = [{BACKUP_CERTIFICATE['name']}])"

                # all backup workflow should execute in single user-restricted connection
                if not UserAccess.SINGLE_USER.value in user_access:
                    cur.execute(f'ALTER DATABASE [{db}] SET single_user WITH ROLLBACK IMMEDIATE')
                
                if should_encrypt:
                    # Need encryptor, otherwise could not backup/restore:
                    prepare_encryptors(cur, BACKUP_CERTIFICATE['name'], BACKUP_CERTIFICATE['subject'], password)
                   
                cur.execute(backup_command)
                wait_busy_cursor(cur)
                cur.execute(f'ALTER DATABASE [{db}] SET multi_user')
                cur.execute(f"RESTORE VERIFYONLY FROM DISK = N'{backup_path}'")
                messages = cur.messages
                # unlikely, but should predict all cases
                if not messages:
                    return f'backup of database {db} could not be verified'
                
                return '' if any('is valid' in str(i[1]).lower() for i in messages) else str(messages)
    except Exception as error:
        return str(error)

def restore_sqlserver_database(host, node, db_name, backup_folder, backup_name, backup_type = BackupType.FULL.value, password = '', restore_date = '', norecovery = True) -> str:
    
    global SYSTEM_DATABASES, BACKUP_CERTIFICATE
    try:
        db = db_name.strip()
        # Skip system databases - restoring it can break the instance
        if db in SYSTEM_DATABASES:
            return (False, 0.0)
        
        # backup may be non-encrypted, but if password provided - ensure encryptors exists, for safety
        may_decrypt = isinstance(password, str) and password != ''
        backup_type_lower = backup_type.strip().lower()
        name = backup_name.strip()
        folder = backup_folder.strip()
        backup_path = f'{folder}\\{name}'
        """
        Only FULL and BULK_LOGGED recovery model allows pitr
        Only full and log backups could be restored to precise datetime
        """
        date = restore_date.strip()
        
        # for restore from snapshot - database <snapshot name> MUST not exist in instance
        if BackupType.SNAPSHOT.value in backup_type_lower:
            is_error, result = drop_database(host, node, backup_name)
            if is_error:
                return is_error, result
        
        conn = connect_to_sqlserver_instance(host['name'], 'master', node['database_user'], node['database_pwd'], node['port'])
        if isinstance(conn, tuple):
            return str(conn[1])
        
        with conn:
            conn.autocommit = True
            with conn.cursor() as cur:
                cur.execute(f"SELECT state_desc, user_access_desc, recovery_model_desc FROM sys.databases WHERE name = N'{db}'")
                res = cur.fetchone()
                database_state = ValidationResult.ERROR.value if not res else res[0].strip().lower()
                user_access = ValidationResult.ERROR.value if not res else res[1].strip().lower()
                recovery_model = RecoveryModel.SIMPLE.value if not res else res[2].strip().lower()
                # that will be used several times, so better as separate expression
                is_simple_recovery = RecoveryModel.SIMPLE.value in recovery_model
                is_full_backup = BackupType.FULL.value in backup_type_lower
                is_log_backup = BackupType.LOG.value in backup_type_lower
                # ensure database exists if restoring non-full backup
                should_db_exist = ValidationResult.ERROR.value in database_state and not is_full_backup
                if should_db_exist:
                    is_error, result = ensure_database_exists(host, node, db_name)
                    if is_error:
                        return is_error, result
               
                if may_decrypt:
                    # encryptors used for backup MUST exist in instance to restore the encrypted backups
                    prepare_encryptors(cur, BACKUP_CERTIFICATE['name'], BACKUP_CERTIFICATE['subject'], password)
                
                # at any times when database may exist - restrict access
                should_restrict_access = DatabaseState.ONLINE.value in database_state
                if should_restrict_access:
                    if UserAccess.SINGLE_USER.value not in user_access:
                        cur.execute(f"ALTER DATABASE [{db}] SET single_user WITH ROLLBACK IMMEDIATE")
                    
                    if not is_log_backup:
                        cur.execute(f"ALTER DATABASE [{db}] SET OFFLINE WITH ROLLBACK IMMEDIATE")
                
                if BackupType.DIFFERENTIAL.value in backup_type_lower and not is_simple_recovery:
                    cur.execute(f"RESTORE DATABASE [{db}] FROM DISK = N'{backup_path}' WITH REPLACE, NORECOVERY")
                    wait_busy_cursor(cur)
                
                if BackupType.LOG.value in backup_type_lower and not is_simple_recovery:
                    if DatabaseState.OFFLINE.value in database_state:
                        # log backups can be restored only on online database
                        cur.execute(f"ALTER DATABASE [{db}] SET ONLINE WITH ROLLBACK IMMEDIATE")
                    if date == '' or BackupType.COPY_ONLY.value in backup_type_lower:
                        cur.execute(f"RESTORE LOG [{db}] FROM DISK = N'{backup_path}' WITH REPLACE, NORECOVERY")
                        wait_busy_cursor(cur)
                    else:
                        try:
                            cur.execute(f"RESTORE LOG [{db}] FROM DISK = N'{backup_path}' WITH STOPAT = N'{date}', NORECOVERY")
                            wait_busy_cursor(cur)
                        except Exception as error_stopat:
                            if 'already rolled forward' not in str(error_stopat).lower():
                                return (True, str(error_stopat))
                            """
                            If already rolled forward, then log has nothing to add,
                            then rebuild broken on error restore checksum by restore WITH REPLACE
                            """
                            cur.execute(f"RESTORE LOG [{db}] FROM DISK = N'{backup_path}' WITH REPLACE, NORECOVERY")
                            wait_busy_cursor(cur)
                
                if BackupType.SNAPSHOT.value in backup_type_lower:
                    # thats not filename, logical name of snapshot
                    cur.execute(f"RESTORE DATABASE [{db}] FROM DATABASE_SNAPSHOT = N'{name}' WITH NORECOVERY")
                    wait_busy_cursor(cur)
                
                if is_full_backup:
                    cur.execute(f"RESTORE DATABASE [{db}] FROM DISK = N'{backup_path}' WITH REPLACE, NORECOVERY")
                    wait_busy_cursor(cur)
                # if recovery mode are simple or no file/differential/log backups left to apply
                if is_simple_recovery or not norecovery:
                    '''
                    docs recommend to do all recovery actions with NORECOVERY,
                    and put recovery state off at the end of recovery
                    '''
                    cur.execute(f'RESTORE DATABASE [{db}] WITH RECOVERY')
                    wait_busy_cursor(cur)
                
                cur.execute(f"SELECT state_desc FROM sys.databases WHERE name = N'{db}'")
                res = cur.fetchone()
                database_state = ValidationResult.ERROR.value if not res else res[0].lower()
                should_unlock_access = DatabaseState.ONLINE.value in database_state
                if should_unlock_access:
                    cur.execute(f"ALTER DATABASE [{db}] SET multi_user")
               
                cur.execute(f"SELECT SUM(mf.size * 8) AS SizeKB FROM sys.master_files mf JOIN sys.databases d ON mf.database_id = d.database_id WHERE DB_NAME(mf.database_id) = N'{db}' GROUP BY mf.database_id")
                res = cur.fetchone()
                return (False, 0.0 if not res or not res[0] else float(res[0]))
    except Exception as error:
        return (True, str(error))

def get_major_version(conn):
    """gets connection to instance and returns major version of instance as tuple"""

    try:
        cur = conn.cursor()
        cur.execute("SELECT CAST(SERVERPROPERTY('productversion') AS nvarchar)")
        major_version = int(cur.fetchone()[0].split(".")[0])
        # cur.execute("SELECT CAST(SERVERPROPERTY('ProductMajorVersion') AS INTEGER)")
        # major_version = cur.fetchone()[0]
    except Exception as error:
        return True, str(error)
    finally:
        cur.close()

    return False, major_version

def get_edition_id(conn) -> tuple:
    """gets connection and returns EditionID of instance as tuple"""

    try:
        cur = conn.cursor()
        cur.execute("SELECT CAST(SERVERPROPERTY('EditionID') AS INTEGER)")
        edition_id = cur.fetchone()[0]
    except Exception as error:
        return True, str(error)
    finally:
        cur.close()

    return False, edition_id

def check_edition_compatibility(source_edition_id: int, destination_edition_id: int) -> str:
    # for database copy
    # 1804890536 = Enterprise
    # 1872460670 = Enterprise Edition: Core-based Licensing
    # 610778273 = Enterprise Evaluation
    # 284895786 = Business Intelligence
    # -2117995310 = Developer
    # -1592396055 = Express
    # -133711905 = Express with Advanced Services
    # -1534726760 = Standard
    ########################
    # entperise -> enterprise, developer => valid
    # enterprise -> standard => invalid
    # developer -> enterprise => valid
    # standard -> standard => valid
    # standard -> enterprise, express => invalid
    # express -> express => valid
    # express -> standard, enterprise => invalid

    compatibility_matrix = {
        1804890536: [1804890536, 1872460670, 610778273, -2117995310],
        1872460670: [1804890536, 1872460670, 610778273, -2117995310],
        -2117995310: [1804890536, 1872460670, 610778273, -2117995310],
        -1534726760: [-1534726760],
        -1592396055: [-1592396055, -133711905],
        -133711905: [-1592396055, -133711905],
        610778273: [1804890536, 1872460670, 610778273, -2117995310],
        284895786: [284895786],
    }

    if source_edition_id in compatibility_matrix:
        if destination_edition_id in compatibility_matrix[source_edition_id]:
            return ValidationResult.VALID.value
        else:
            return ValidationResult.INVALID.value
    else:
        return ValidationResult.ERROR.value

def register_sqlserver_instance_task(data, task_id):
    """register SQL Server instance to app"""

    task = 'register sqlserver instance task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()
    
    try:
        host = json.loads(data['host_data'])
        node = json.loads(data['node_data'])
        db_name = 'master'
        minimum_version = '16'
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'checking register prerequisites'
    duration_step_start = time.perf_counter()

    sql_conn = connect_to_sqlserver_instance(host['name'], db_name, node['database_user'], node['database_pwd'], node['port'])
    if isinstance(sql_conn, tuple):
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(sql_conn[1]),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(sql_conn[1]),1)
        return
    
    try:
        is_error, major_version = get_major_version(sql_conn)
        if is_error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(major_version),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(major_version),1)
            return
    
        if major_version < int(minimum_version):
            error = f'123Cluster app does not support SQL Server versions below {minimum_version}'
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return
        
        sql_cur = sql_conn.cursor()
        sql_cur.execute('SELECT @@SERVICENAME')
        res = sql_cur.fetchone()
        # provide fail-safe default where we can
        service_name = res[0] if res and res[0] else 'MSSQLSERVER'
        sql_cur.execute("SELECT SERVERPROPERTY('ishadrenabled')")
        res = sql_cur.fetchone()
        standalone = res[0] if res and res[0] else 0
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'registering instance databases'
    duration_step_start = time.perf_counter()

    try:
        with get_postgres_connection() as conn:
            with conn.cursor(cursor_factory=RealDictCursor) as cur:
                cur.execute("""SELECT DISTINCT ON (name) id, name FROM "clusterCore_c123_database" WHERE node_id=%s""", (node['id'],))
                db_list_from_app = cur.fetchall()
                sql_cur.execute("SELECT name FROM sys.databases")
                db_list_from_node = [row[0] for row in sql_cur.fetchall()]
                for db in db_list_from_app:
                    db_name = db['name'].strip()
                    if db_name in db_list_from_node:
                        sql_cur.execute(f"SELECT SUM(mf.size * 8) AS SizeKB FROM sys.master_files mf JOIN sys.databases d ON mf.database_id = d.database_id WHERE DB_NAME(mf.database_id) = N'{db_name}' GROUP BY mf.database_id")
                        res = sql_cur.fetchone()
                        db_size = float(res[0]) if res and res[0] else 0.0
                        cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, "size"=%s, ind_started='true', ind_deleted='false' WHERE id=%s""", (Severity.OK.value, Status.OK.value, db_size, db['id'],))
                        conn.commit()
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = "updating 123cluster app metadata"
    duration_step_start = time.perf_counter()

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity='%s', status='%s', ind_started='true', ind_deleted='false', version=%s, ex_params = ex_params || '{"service_name":\"%s\", "standalone":\"%s\"}' WHERE id=%s""" % (Severity.OK.value, Status.OK.value, major_version, service_name, standalone, node['id']))
                conn.commit()
                # cur.execute("""SELECT app_123monitor.create_node_monitor(%s,%s,%s)""", [org_id,node['id'],created_by])
                # conn.commit()
                # cur.execute("""SELECT app_123monitor.create_host_monitor(%s,%s,%s,%s)""", [org_id,host['id'],created_by,'WIN'])
                # conn.commit()
        
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    except Exception as error:
       status = Severity.WARNING.value
       duration_step_end = time.perf_counter() - duration_step_start
       logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def unregister_sqlserver_instance_task(data, task_id):
    """unregister SQL Server instance from app"""
    
    task = 'unregister sqlserver instance task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()
    
    try:
        node = json.loads(data['node_data'])
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'unregistering sqlserver instance from app'
    duration_step_start = time.perf_counter()

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status='deleted', ind_started='false', ind_deleted='true' WHERE id=%s""", [Severity.NOT_OK.value, node['id']])
                cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status='deleted', ind_deleted='true' WHERE node_id=ANY(%s)""", (Severity.NOT_OK.value, node["id"],))
                conn.commit()
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def backup_sqlserver_instance_task(data, task_id):
    # Can`t backup the instance, but can backup all databases in separate backup files and made archive from that files
    
    task = 'backup sqlserver instance task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()
    
    try:
        global SYSTEM_DATABASES
        global BACKUP_CERTIFICATE
        host = json.loads(data['host_data'])
        node = json.loads(data['node_data'])
        backup_host = json.loads(data['backup_host'])
        backup = json.loads(data['backup_data'])
        backup_path = backup['ex_params'].get("backup_path")
        backup_name = backup['name']
        compress = backup['ex_params'].get('compress')
        encryption = backup['ex_params'].get('encryption')
        backslash_backup_path = backup_path.replace('/', '\\')
        tmp_backup_path = f"{TASK_TMP_FILES_DIR_WIN}{backslash_backup_path}"
        #since copy folder to linux - should transfer to parent path
        backup_root = os.path.dirname(backup_path)
        is_sched = backup["is_sched"]
        backup_status = backup['status']
        # tar is available by default in windows 10/11 only
        full_archive = f"{tmp_backup_path}\\{backup_name}@full.tar.gz"
        log_archive = f"{tmp_backup_path}\\{backup_name}@log.tar.gz"
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    step = 'backing up the databases of the node'
    duration_step_start = time.perf_counter()

    sql_conn = connect_to_sqlserver_instance(host['name'], 'master', node['database_user'], node['database_pwd'], node['port'])
    if isinstance(sql_conn, tuple):
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(sql_conn[1]),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(sql_conn[1]),1)
        return
    else:
        try:
            with sql_conn:
                with sql_conn.cursor() as sql_cur:
                    sql_cur.execute("SELECT name FROM sys.databases WHERE name NOT IN ({})".format(','.join("?"*len(SYSTEM_DATABASES))), SYSTEM_DATABASES)
                    database_names = [i[0] for i in sql_cur.fetchall()]
                    jsn = json.loads(remote_command_win(host, f'if not exist "{tmp_backup_path}" mkdir "{tmp_backup_path}"', True), strict=False)
                    if jsn['stderr'] != '':
                        status = Severity.NOT_OK.value
                        duration_step_end = time.perf_counter() - duration_step_start
                        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        step = 'null'
                        status = Severity.NOT_OK.value
                        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                        return
                    
                    # backup error yellow - single database failure does not mean we dont have good backups
                    last_known_err = ''
                    for db_name in database_names:
                        if is_sched:
                            with get_postgres_connection() as conn:
                                with conn.cursor() as cur:
                                    # we want it as relevant as possible
                                    cur.execute("""SELECT status FROM "clusterCore_c123_inventory" WHERE id=%s""", (backup['id'],))
                                    res = cur.fetchone()
                                    if res and res[0]:
                                        backup_status = res[0]
                                    
                            # until we have at least one successful backup run - default to full backup
                            if not backup_status:
                                last_known_err = backup_sqlserver_database(host, node, db_name, tmp_backup_path, db_name, encryption=encryption, password=BACKUP_CERTIFICATE['password'])
                                if last_known_err:
                                    logging.error(f"{task}: {task_id}: {step}: {db_name}: {last_known_err}")
                            # when have at least one successful backup run - default to log backup
                            else:
                                last_known_err = backup_sqlserver_database(host, node, db_name, tmp_backup_path, db_name, BackupType.LOG.value, encryption=encryption, password=BACKUP_CERTIFICATE['password'])
                                if last_known_err:
                                    logging.error(f"{task}: {task_id}: {step}: {db_name}: {last_known_err}")
                        else:
                            last_known_err = backup_sqlserver_database(host, node, db_name, tmp_backup_path, db_name, encryption=encryption, password=BACKUP_CERTIFICATE['password'])
                            if last_known_err:
                                logging.error(f"{task}: {task_id}: {step}: {db_name}: {last_known_err}")
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return
    
    if compress:
        try:
            jsn = json.loads(remote_command_win(host, f'tar --exclude "*.tar.gz" -czf "{log_archive if is_sched and backup_status else full_archive}" -C "{tmp_backup_path}" . ; Get-ChildItem -Path "{tmp_backup_path}" -File | Where-Object {{ $_.Extension -ne ".gz" }} | Remove-Item -Force', True, "ps"), strict=False)
            if jsn['excode'] != 0:
                # throw red step, could not fulfill purpose of compressed
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return
    
    status = Severity.OK.value if not last_known_err else 'yellow'
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by) if status == 'green' else logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,'',last_known_err,1)
    
    step = 'copying backup to backup host'
    duration_step_start = time.perf_counter()

    try:
        jsn = json.loads(remote_command(backup_host, f"sudo mkdir -p '{backup_path}'", True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        modified_date = get_date_modified_folder(backup_host, backup_path)
        if isinstance(modified_date, tuple):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(modified_date[1]),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(modified_date[1]),1)
            return
        
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = ex_params || '{"modified_date":\"%s\"}' WHERE id=%s""" % (modified_date, backup['id']))
                conn.commit()
        
        # rather slow, takes a minute to copy 30 MB file...
        is_error, result = copy_files_linux_win(backup_host, host, backup_root, tmp_backup_path, True)
        if is_error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,result,1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,result,1)
            return
        
        jsn = json.loads(remote_command(backup_host, f"sudo du -bs '{backup_path}' | awk '{{print $1}}'", True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        backup_size = float(jsn['stdout']) / 1024.0

        # remote_command_win(host, f'rd /s /q "{tmp_backup_path}"', True)
        
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = jsonb_set(ex_params, '{{size}}', '{0}', true), status='true' WHERE id={1}""".format(backup_size,backup['id']))
                cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = jsonb_set(ex_params, '{{duration}}', '{0}', true) WHERE id={1}""".format(duration_task_end,backup['id']))
                conn.commit()
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def create_sqlserver_database_task (data, task_id):
    """
    Create SQL Server database.
    3 users => 1.owner 2.all dml (insert update delete - not drop or alter) 3.select
    """
        
    task = 'create sqlserver database task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()
    
    try:
        host = json.loads(data['host_data'])
        node = json.loads(data['node_data'])
        db = json.loads(data['db'])
        db_name = db['name']
        recovery_model = data['recovery_model']
        schema_name = f"{db_name}_schema"
        owner_user_name = f"{db_name}_owner_user"
        write_user_name = f"{db_name}_write_user"
        read_user_name = f"{db_name}_read_user"
        
        owner_user_pwd = generate_complex_password()
        write_user_pwd = generate_complex_password()
        read_user_pwd = generate_complex_password()
        create_users_output = f'USERNAME = {owner_user_name}\nPASSWORD = {owner_user_pwd}\nUSERNAME = {read_user_name}\nPASSWORD = {read_user_pwd}\nUSERNAME = {write_user_name}\nPASSWORD = {write_user_pwd}'
        create_user_queries = [
            f'USE [{db_name}]',
            f'''IF NOT EXISTS (SELECT * FROM sys.server_principals WHERE name = '{owner_user_name}') BEGIN CREATE LOGIN [{owner_user_name}] WITH PASSWORD = '{owner_user_pwd}', DEFAULT_DATABASE=[{db_name}] END''',
            f'''IF NOT EXISTS (SELECT * FROM sys.server_principals WHERE name = '{write_user_name}') BEGIN CREATE LOGIN [{write_user_name}] WITH PASSWORD = '{write_user_pwd}', DEFAULT_DATABASE=[{db_name}] END''',
            f'''IF NOT EXISTS (SELECT * FROM sys.server_principals WHERE name = '{read_user_name}') BEGIN CREATE LOGIN [{read_user_name}] WITH PASSWORD = '{read_user_pwd}', DEFAULT_DATABASE=[{db_name}] END''', 
            f"CREATE SCHEMA [{schema_name}]",
            f"CREATE USER [{owner_user_name}] FOR LOGIN [{owner_user_name}] WITH DEFAULT_SCHEMA=[{schema_name}]",
            f"ALTER ROLE db_owner ADD MEMBER [{owner_user_name}]",
            f"CREATE USER [{write_user_name}] FOR LOGIN [{write_user_name}] WITH DEFAULT_SCHEMA=[{schema_name}]",
            f"GRANT SELECT, INSERT, UPDATE ON SCHEMA::[{schema_name}] TO [{write_user_name}]",
            f"CREATE USER [{read_user_name}] FOR LOGIN [{read_user_name}] WITH DEFAULT_SCHEMA=[{schema_name}]",
            f"GRANT SELECT ON SCHEMA::[{schema_name}] TO [{read_user_name}]",
            f"SELECT ROUND(SUM(size * 8 / 1024.0), 2) AS [Size_KB] FROM sys.databases AS db JOIN sys.master_files AS mf ON db.database_id = mf.database_id WHERE db.name ='{db_name}'"
        ]
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'creating database'
    duration_step_start = time.perf_counter()

    is_error, result = ensure_database_exists(host, node, db_name, recovery_model)
    if is_error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,result,1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,result,1)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'creating database users'
    duration_step_start = time.perf_counter()

    try:
        sql_conn = connect_to_sqlserver_instance(host['name'], 'master', node['database_user'], node['database_pwd'], node['port'])
        if isinstance(sql_conn, tuple):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(sql_conn[1]),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(sql_conn[1]),1)
            return
        
        with sql_conn:
            sql_conn.autocommit = True
            with sql_conn.cursor() as sql_cur:
                for q in create_user_queries:
                    sql_cur.execute(q)
                
                res = sql_cur.fetchone()
                db_size = float(res[0]) if res and res[0] else 0.0
                sql_cur.execute("SELECT SUM(mf.size * 8) AS SizeKB FROM sys.master_files mf JOIN sys.databases d ON mf.database_id = d.database_id WHERE mf.database_id > 4 GROUP BY mf.database_id")
                res = sql_cur.fetchone()
                node_size = res[0] if res and res[0] else 0.0
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,create_users_output,None,0)
    
    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, size=%s, ind_started='true', ind_deleted='false' WHERE id=%s""", [Severity.OK.value, Status.OK.value, db_size, db['id']])
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true', ind_deleted='false', size=%s WHERE id=%s""", [Severity.OK.value, Status.OK.value, node_size, node['id']])
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def start_sqlserver_database_task(data, task_id):
    """start SQL Server database"""
    
    task = 'start sqlserver database task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()
    
    try:
        host = json.loads(data['host_data'])
        node = json.loads(data['node_data'])
        db = json.loads(data['db'])
        db_name = db['name']
        db_id = db['id']
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    step = 'starting the database'
    duration_step_start = time.perf_counter()
    
    sql_conn = connect_to_sqlserver_instance(host['name'], 'master', node['database_user'], node['database_pwd'], node['port'])
    if isinstance(sql_conn, tuple):
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(sql_conn[1]),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(sql_conn[1]),1)
        return
    
    try:
        sql_conn.autocommit = True
        sql_cur = sql_conn.cursor()
        sql_cur.execute("USE master")
        sql_cur.execute("ALTER DATABASE [?] SET ONLINE", db_name)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    finally:
        sql_cur.close()
        sql_conn.close()
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='true', ind_deleted='false' WHERE id=%s""", (Severity.OK.value, Status.OK.value, db_id))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)
 
def restart_sqlserver_database_task(data, task_id):
    """restart SQL Server database"""
    
    task = 'restart sqlserver database task'
    
    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()
    
    try:
        host = json.loads(data['host_data'])
        node = json.loads(data['node_data'])
        db = json.loads(data['db'])
        db_name = db['name']
        db_id = db['id']
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    step = 'restarting the database'
    duration_step_start = time.perf_counter()
    
    sql_conn = connect_to_sqlserver_instance(host['name'], 'master', node['database_user'], node['database_pwd'], node['port'])
    if isinstance(sql_conn, tuple):
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(sql_conn[1]),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(sql_conn[1]),1)
        return
    
    try:
        sql_conn.autocommit = True
        sql_cur = sql_conn.cursor()
        sql_cur.execute("ALTER DATABASE [{}] SET SINGLE_USER WITH ROLLBACK IMMEDIATE".format(db_name))
        sql_cur.execute("ALTER DATABASE [{}] SET MULTI_USER".format(db_name))
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    finally:
        sql_cur.close()
        sql_conn.close()
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='true', ind_deleted='false' WHERE id=%s""", (Severity.OK.value, Status.OK.value, db_id,))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def stop_sqlserver_database_task(data, task_id):
    """stop SQL Server database"""
    
    task = 'stop sqlserver database task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()
    
    try:
        host = json.loads(data['host_data'])
        node = json.loads(data['node_data'])
        db = json.loads(data['db'])
        db_name = db['name']
        db_id = db['id']
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    step = 'stopping the database'
    duration_step_start = time.perf_counter()
    
    sql_conn = connect_to_sqlserver_instance(host['name'], 'master', node['database_user'], node['database_pwd'], node['port'])
    if isinstance(sql_conn, tuple):
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(sql_conn[1]),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(sql_conn[1]),1)
        return
    
    try:
        sql_conn.autocommit = True
        sql_cur = sql_conn.cursor()
        # MUST be always used WITH ROLLBACK IMMEDIATE, to prevent deadlock
        sql_cur.execute(f"ALTER DATABASE [{db_name}] SET OFFLINE WITH ROLLBACK IMMEDIATE")
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    finally:
        sql_cur.close()
        sql_conn.close()
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status='stopped', ind_started='false', ind_deleted='false' WHERE id=%s""", (Severity.NOT_OK.value, db_id))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def drop_sqlserver_database_task(data, task_id):
    """drop SQL Server database"""
    
    task = 'drop sqlserver database task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()
    
    try:
        host = json.loads(data['host_data'])
        node = json.loads(data['node_data'])
        db = json.loads(data['db_data'])
        db_name = db['name']
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    step = 'dropping the database'
    duration_step_start = time.perf_counter()
    
    is_error, result = drop_database(host, node, db_name)
    if is_error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,result,1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,result,1)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='false', ind_deleted='true' WHERE id=%s""", [Severity.NOT_OK.value, Status.DROPPED.value, db['id']])
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true', ind_deleted='false', size=%s WHERE id=%s""", [Severity.OK.value, Status.OK.value, result, node['id']])
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def rename_sqlserver_database_task(data, task_id):
    """rename SQL Server database"""
    
    task = 'rename sqlserver database task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()
    
    try:
        host = json.loads(data['host_data'])
        node = json.loads(data['node_data'])
        db = json.loads(data['db'])
        db_name = db['name']
        db_id = db['id']
        new_db_name = data['rename'] #follow oracle patterns
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    step = 'renaming the database'
    duration_step_start = time.perf_counter()

    sql_conn = connect_to_sqlserver_instance(host['name'], 'master', node['database_user'], node['database_pwd'], node['port'])
    if isinstance(sql_conn, tuple):
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(sql_conn[1]),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(sql_conn[1]),1)
        return
    
    try:
        sql_conn.autocommit = True
        sql_cur = sql_conn.cursor()
        sql_cur.execute(f"ALTER DATABASE [{db_name}] MODIFY NAME = {new_db_name}")
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    finally:
        sql_cur.close()
        sql_conn.close()
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_database" SET name=%s, severity=%s, status=%s, ind_started='true', ind_deleted='false' WHERE id=%s""", [Severity.OK.value, Status.OK.value, new_db_name, db_id])
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def stop_sqlserver_instance_task(data, task_id):
    """stop SQL Server instance"""
    
    task = 'stop sqlserver instance task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()
    
    try:
        host = json.loads(data['host_data'])
        node = json.loads(data['node_data'])
        service_name = node['ex_params'].get('service_name')
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    step = 'stopping the instance'
    duration_step_start = time.perf_counter()
    
    try:
        jsn = json.loads(remote_command_win(host, f'net stop {service_name} /y', True, 'ps'), strict=False)
        if jsn['excode'] != 0 and 'not started' not in jsn['stderr'].lower():
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status='stopped', ind_started='false', ind_deleted='false' WHERE id=%s""", [Severity.OK.value, node['id']])
                cur.execute("""UPDATE "clusterCore_c123_database" SET status=%s, ind_started='false' WHERE node_id=%s AND ind_deleted='false'""", (Status.OK.value, node['id'],))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def start_sqlserver_instance_task(data, task_id):
    """start SQL Server instance"""
    
    task = 'start sqlserver instance task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()
    
    try:
        host = json.loads(data['host_data'])
        node = json.loads(data['node_data'])
        service_name = node['ex_params'].get('service_name')
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    step = 'starting the instance'
    duration_step_start = time.perf_counter()

    try:
        jsn = json.loads(remote_command_win(host, f'net start {service_name}', True, 'ps'), strict=False)
        if int(jsn['excode']) != 0:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true', ind_deleted='false' WHERE id=%s""", [Severity.OK.value, Status.OK.value, node['id']])
                cur.execute("""UPDATE "clusterCore_c123_database" SET status=%s, ind_started='true' WHERE node_id=%s AND ind_deleted='false'""", (Status.OK.value, node['id'],))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def restart_sqlserver_instance_task(data, task_id):
    """restart SQL Server instance"""
    
    task = 'restart sqlserver instance task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()
    
    try:
        host = json.loads(data['host_data'])
        node = json.loads(data['node_data'])
        service_name = node['ex_params'].get('service_name')
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    step = 'restarting the instance'
    duration_step_start = time.perf_counter()

    try:
        commands = [
            f'net stop {service_name} /y',
            f'net start {service_name}'
        ]
        for command in commands:
            jsn = json.loads(remote_command_win(host, command, True, 'ps'), strict=False)
            if int(jsn['excode']) != 0 and not ('net stop' in command and 'not started' in jsn['stderr'].lower()):
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true', ind_deleted='false' WHERE id=%s""", [Severity.OK.value, Status.OK.value, node['id']])
                cur.execute("""UPDATE "clusterCore_c123_database" SET status=%s, ind_started='true' WHERE node_id=%s AND ind_deleted='false'""", (Status.OK.value, node['id'],))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def copy_sqlserver_database_task(data, task_id):
    """copy SQL Server database in same instance or to another server"""

    task = 'copy sqlserver database task'
    
    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()
    
    try:
        host_source = json.loads(data['host_source'])
        node_source = json.loads(data['node_source'])
        host_target = json.loads(data['host_target'])
        node_target = json.loads(data['node_target'])
        db_option = data['db_option']
        db_exists = data['db_exists']
        # copy_option = data['copy_option']
        copy_name = data.get('copy_name')
        db_source = json.loads(data['db_source'])
        db_source_name = db_source['name']
        db_target = json.loads(data['db_target'])
        db_source_id = db_source['id']
        db_target_id = db_target['id']
        override = db_option == 'override'
        new_db_name = copy_name if isinstance(copy_name, str) and len(copy_name.strip()) > 0  else db_target["name"]
        #each task that working with files, should have different storage folders
        target_tmp_path = f'{TASK_TMP_FILES_DIR_WIN}\\{node_source["name"].replace(".", "_").replace(":", "_")}\\copy'
        target_tmp_backup = f'{target_tmp_path}\\{new_db_name}_tmp.bak'
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'checking copy prerequisites'
    duration_step_start = time.perf_counter()

    sql_source_conn = connect_to_sqlserver_instance(host_source['name'], 'master', node_source['database_user'], node_source['database_pwd'], node_source['port'])
    if isinstance(sql_source_conn, tuple):
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(sql_source_conn[1]),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(sql_source_conn[1]),1)
        return

    try:
        sql_source_conn.autocommit = True
        sql_source_cur = sql_source_conn.cursor()
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    if node_source['name'] == node_target['name'] and node_source['port'] == node_target['port']:
        sql_target_conn = sql_source_conn
        sql_target_cur = sql_source_cur
    else:
        sql_target_conn = connect_to_sqlserver_instance(host_target['name'], 'master', node_target['database_user'], node_target['database_pwd'], node_target['port'])
        if isinstance(sql_source_conn, tuple):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(sql_source_conn[1]),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(sql_source_conn[1]),1)
            return
       
        try:
            sql_target_conn.autocommit = True
            sql_target_cur = sql_target_conn.cursor()
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return
    
    is_error, source_major_version = get_major_version(sql_source_conn)
    if is_error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(source_major_version),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(source_major_version),1)
        return

    is_error, destination_major_version = get_major_version(sql_target_conn)
    if is_error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(destination_major_version),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(destination_major_version),1)
        return
    
    if int(destination_major_version) < int(source_major_version):
        error = f"Database copy could not be performed.\nDestination instance version {destination_major_version} are less than source instance version {source_major_version}"
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,1)
        return
    
    is_error, source_edition_id = get_edition_id(sql_source_conn)
    if is_error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(source_edition_id),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(source_edition_id),1)
        return

    is_error, destination_edition_id = get_edition_id(sql_target_conn)
    if is_error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(destination_edition_id),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(destination_edition_id),1)
        return
    
    res = check_edition_compatibility(source_edition_id, destination_edition_id)
    if res != ValidationResult.VALID.value:
        if res == ValidationResult.ERROR.value:
            logging.error(f"{task}: unsupported source instance edition {source_edition_id} detected in task {task_id}")
        error = f"Database copy could not be performed.Destination instance edition ID {destination_edition_id} are incompatible with source instance edition ID {source_edition_id}."
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,1)
        return
    
    try:
        sql_source_cur.execute(f"SELECT recovery_model_desc FROM sys.databases WHERE name = N'{db_source_name}'")
        res = sql_source_cur.fetchone()
        recovery_model = RecoveryModel.SIMPLE.name if not res else res[0]
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'creating the temporary backup of database'
    duration_step_start = time.perf_counter()

    try:
        jsn = json.loads(remote_command_win(host_source, f'if not exist "{target_tmp_path}" mkdir "{target_tmp_path}"', True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        #this backup are temporary - metadata does not needed
        sql_source_cur.execute(f'''BACKUP DATABASE [{db_source_name}] TO DISK = '{target_tmp_backup}' WITH FORMAT, INIT''')
        wait_busy_cursor(sql_source_cur)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if host_source['name'] != host_target['name']:
        step = 'copy backup file to other host'
        duration_step_start = time.perf_counter()
        
        try:
            jsn = json.loads(remote_command_win(host_target, f'if not exist "{target_tmp_path}" mkdir "{target_tmp_path}"', True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            command = f'''
        $passwd = convertto-securestring -AsPlainText -Force -String "{host_target["password"]}"
        $cred = new-object -typename System.Management.Automation.PSCredential -argumentlist "administrator@robert",$passwd
        $spn = New-PSSessionOption -IncludePortInSPN
        $s = New-PSSession -ComputerName {host_target["name"]} -Port 5985 -Credential $cred -SessionOption $spn
        Invoke-Command -ScriptBlock {{Copy-Item -Path '{target_tmp_backup}' -Destination '{target_tmp_backup}' -ToSession $s}}'''
            jsn = json.loads(remote_command_win(host_source, command, True, "ps"))
            if jsn['excode'] != 0:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
            
            remote_command_win(host_source, f'rd /s /q "{target_tmp_path}"', True)
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return
        
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'copying the database'
    duration_step_start = time.perf_counter()

    try:
        if db_exists and override:
            is_error, result = drop_database(host_target, node_target, new_db_name)
            if is_error:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,result,1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,result,1)
                return

        is_error, result = ensure_database_exists(host_target, node_target, new_db_name)
        if is_error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,result,1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,result,1)
            return
        
        # not checking db state, since here we should copy to fresh db, if not - something wrong
        sql_target_cur.execute(f"ALTER DATABASE [{new_db_name}] SET single_user WITH ROLLBACK IMMEDIATE")
        sql_target_cur.execute(f"ALTER DATABASE [{new_db_name}] SET OFFLINE WITH ROLLBACK IMMEDIATE")
        if db_source_name == new_db_name:
            # then we can use safe classic approach
            restore_query = f"RESTORE DATABASE [{new_db_name}] FROM DISK = N'{target_tmp_backup}' WITH REPLACE, NORECOVERY"
        # our manual copy
        else:
            # fetching backup file list
            sql_target_cur.execute(f"RESTORE FILELISTONLY FROM DISK = N'{target_tmp_backup}'")
            file_infos = sql_target_cur.fetchall()

            # Manually restoring every (!) backup file to other database
            file_list = ''
            for file_info in file_infos:
                physical_name = file_info.PhysicalName.replace(db_source_name, new_db_name)
                file_list += f"MOVE N'{file_info.LogicalName}' TO N'{physical_name}', "
                
            restore_query = f"RESTORE DATABASE [{new_db_name}] FROM DISK = N'{target_tmp_backup}' WITH {file_list} REPLACE, NORECOVERY"
        
        sql_target_cur.execute(restore_query)
        wait_busy_cursor(sql_target_cur)
        sql_target_cur.execute(f"RESTORE DATABASE [{new_db_name}] WITH RECOVERY")
        wait_busy_cursor(sql_target_cur)
        sql_target_cur.execute(f"ALTER DATABASE [{new_db_name}] SET RECOVERY {recovery_model}")
        sql_target_cur.execute(f"ALTER DATABASE [{new_db_name}] SET multi_user")
        remote_command_win(host_target, f'rd /s /q "{target_tmp_path}"', True)
        
        queries = [
            # database backup does not backup logins needed for users and schema, but it may exist if copy to same instance
            f'''IF NOT EXISTS (SELECT * FROM sys.server_principals WHERE name = '{db_source['name']}_owner_user') BEGIN CREATE LOGIN [{db_source['name']}_owner_user] WITH PASSWORD = 'fR67#sL*2!', DEFAULT_DATABASE=[{db_target['name']}] END''',
            f'''IF NOT EXISTS (SELECT * FROM sys.server_principals WHERE name = '{db_source['name']}_write_user') BEGIN CREATE LOGIN [{db_source['name']}_write_user] WITH PASSWORD = 'J$8&n@9!qB', DEFAULT_DATABASE=[{db_target['name']}] END''',
            f'''IF NOT EXISTS (SELECT * FROM sys.server_principals WHERE name = '{db_source['name']}_read_user') BEGIN CREATE LOGIN [{db_source['name']}_read_user] WITH PASSWORD = 'P#3s$L@!0y', DEFAULT_DATABASE=[{db_target['name']}] END''',
            f"USE [{new_db_name}]",
            f"EXEC sp_change_users_login 'Auto_Fix', '{db_source_name}_owner_user'",
            f"EXEC sp_change_users_login 'Auto_Fix', '{db_source_name}_write_user'",
            f"EXEC sp_change_users_login 'Auto_Fix', '{db_source_name}_read_user'",
        ]
        for q in queries:
            sql_target_cur.execute(q)
        
        sql_target_cur.execute("SELECT SUM(mf.size * 8) AS SizeKB FROM sys.master_files mf JOIN sys.databases d ON mf.database_id = d.database_id WHERE mf.database_id > 4 GROUP BY mf.database_id")
        res = sql_target_cur.fetchone()
        node_size = res[0] if res and res[0] else 0.0
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    finally:
        sql_source_cur.close()
        sql_source_conn.close()
        if sql_target_conn != sql_source_conn:
            sql_target_cur.close()
            sql_target_conn.close()

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='true', ind_deleted='false' WHERE id=%s""", (Severity.OK.value, Status.OK.value, db_target_id,))
                cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='true', ind_deleted='false' WHERE id=%s""", (Severity.OK.value, Status.OK.value, db_source_id,))
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true', ind_deleted='false', size=%s WHERE id=%s""", [Severity.OK.value, Status.OK.value, node_size, node_target['id']])
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def change_sqlserver_database_size_task(data, task_id):
    """Change the SQL Server database`s files size"""
    
    task = 'change sqlserver database size task'
    
    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()
    
    try:
        host = json.loads(data['host_data'])
        node = json.loads(data['node_data'])
        db = json.loads(data['db_data'])
        db_name = db['name']
        db_id = db['id']
        new_size = json.loads(data["new_file_size"])
        change_the_log_file = json.loads(data["new_file_size"])
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)	
    
    step = 'changing the database size'
    duration_step_start = time.perf_counter()
    
    sql_conn = connect_to_sqlserver_instance(host['name'], db_name, node['database_user'], node['database_pwd'], node['port'])
    if isinstance(sql_conn, tuple):
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(sql_conn[1]),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(sql_conn[1]),1)
        return
    
    try:
        sql_conn.autocommit = True
        sql_cur = sql_conn.cursor()
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    if change_the_log_file is True:
        file_id = '2'
    else:
        file_id = '1'
        try:
            sql_cur.execute(f'USE {db_name}')
            sql_cur.execute(f'SELECT FILE_NAME({file_id})')
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return

        logical_name = sql_cur.fetchone()[0]

    try:
        sql_cur.execute('USE master')
        sql_cur.execute(f'ALTER DATABASE [{db_name}] MODIFY FILE(NAME = {logical_name}, SIZE = {new_size}KB)')
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    step = 'getting new database size'
    duration_step_start = time.perf_counter()
    
    try:
        sql_cur.execute(f'''USE {db_name}
            SELECT SUM(size * 8) / 1024 AS size_kb
            FROM sys.master_files
            WHERE database_id = DB_ID()
            GROUP BY database_id
           ''')
        res = sql_cur.fetchone()
        database_size = float(res[0]) if res and res[0] else 0.0
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    finally:
        sql_cur.close()
        sql_conn.close()
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_database" SET size=%s severity=%s, status=%s, ind_started='true', ind_deleted='false' WHERE id=%s""", (database_size, Severity.OK.value, Status.OK.value, db_id))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def backup_sqlserver_database_task(data, task_id):
    """backup sql server database"""
    """not used in app for now (for database are exports, not backups, by general agreement)"""
    
    task = 'backup sqlserver database task'
    
    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()
    
    try:
        global BACKUP_CERTIFICATE
        host = json.loads(data['host_data'])
        node = json.loads(data['node_data'])
        backup_host = json.loads(data['backup_host'])
        db_name = json.loads(data['db_name'])
        backup = json.loads(data['backup_data'])
        backup_path = backup['ex_params'].get("backup_path")
        backup_name = backup['name']
        compress = backup['ex_params'].get("compress")
        encryption = backup['ex_params'].get("encryption")
        backslash_backup_path = backup_path.replace('/', '\\')
        tmp_backup_path = f"{TASK_TMP_FILES_DIR_WIN}{backslash_backup_path}"
        #since copy folder to linux - should transfer to parent path
        backup_root = os.path.dirname(backup_path)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    # if is_sched:
    #     job = json.loads(data['job_data'])
    #     # list of days for example can be [3, 25, 26, 28] 
    #     job_month_days = job['ex_params'].get('job_month_days')
    #     if job_month_days:
    #         job_condition_command = f'''
    #         IF DAY(GETDATE()) = ANY({job_month_days})
    #         BAGIN
    #         '''
    #     job_week_days = job['ex_params'].get('job_week_days') # example: [1, 2, 4, 8, 16, 32, 64] - every day of the week (1 - Sunday 2 - Monday 4 - Tuesday ... 64 - Saturday)

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    step = 'backing up the database'
    duration_step_start = time.perf_counter()

    try:
        jsn = json.loads(remote_command_win(host, f'if not exist "{tmp_backup_path}" mkdir "{tmp_backup_path}"', True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    """
    Developers memo by Dmitriy: 
    If dont check recovery mode of database (and why dont check, 
    this task dont need to support pitr?) - backup should be full, for compatibility.
    """
    error = backup_sqlserver_database(host, node, db_name,tmp_backup_path, backup_name, encryption=encryption, password=BACKUP_CERTIFICATE['password'])
    if error != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    if compress:
        try:    
            # tar is avalable by default in windows 10 and 11 only
            commands = [
                f'''tar -zcf "{tmp_backup_path}\\{backup_name}.tar.gz" -C "{tmp_backup_path}" {backup_name}.bak''',
                f'del /f "{tmp_backup_path}\\{backup_name}.bak"'
            ]
            for command in commands:
                jsn = json.loads(remote_command_win(host, command, True), strict=False)
                if jsn['stderr'] != '':
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                    return
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
            return
    
    step = 'copying backup to backup host'
    duration_step_start = time.perf_counter()

    try:
        jsn = json.loads(remote_command(backup_host, f"sudo mkdir -p '{backup_path}'", True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        modified_date = get_date_modified_folder(backup_host, backup_path)
        if isinstance(modified_date, tuple):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(modified_date[1]),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(modified_date[1]),1)
            return
        
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = ex_params || '{"modified_date":\"%s\"}' WHERE id=%s""" % (modified_date, backup['id']))
                conn.commit()
        
        is_error, result = copy_files_linux_win(backup_host, host, backup_root, tmp_backup_path, True)
        if is_error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,result,1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,result,1)
            return
        
        jsn = json.loads(remote_command(backup_host, f"sudo du -bs '{backup_path}' | awk '{{print $1}}'", True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        backup_size = float(jsn['stdout']) / 1024.0

        remote_command_win(host, f'rd /s /q "{tmp_backup_path}"', True)
        
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = jsonb_set(ex_params, '{{size}}', '{0}', true), status='true' WHERE id={1}""".format(backup_size,backup['id']))
                    cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = jsonb_set(ex_params, '{{duration}}', '{0}', true) WHERE id={1}""".format(duration_task_end,backup['id']))
                    conn.commit()
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def restore_sqlserver_database_from_backup_task(data, task_id):
    """Task for restoring the database from backup"""
    """not used in app for now (for database exports, not backups, by general agreement)"""
    
    task = 'restore sqlserver database from backup task'
    
    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()
    
    try:
        global BACKUP_CERTIFICATE
        host = json.loads(data['host_data'])
        backup_host = json.loads(data['backup_host'])
        node = json.loads(data['node_data'])
        db_data = json.loads(data['db_data'])
        db_name = db_data["name"]
        db_id = db_data['id']
        backup = json.loads(data['backup_data'])
        backup_path = backup['ex_params'].get("backup_path")
        backup_name = backup['name']
        compress = backup['ex_params'].get("compress")
        backup_type = backup["type"]
        backup_options = backup['ex_params'].get("backup_options")
        """
        From Dmitriy:
        TODO Why restore_sqlserver_database_from_backup_task have pitr inputs,
        but backup_sqlserver_database_task dont have?
        """
        restore_option = data['restore_option'].strip().lower()

        if restore_option in ['custom', 'latest']:
            restore_date = data['date']
        else:
            restore_date = ''
        
        backslash_backup_path = backup_path.replace('/', '\\')
        tmp_backup_path = f"{TASK_TMP_FILES_DIR_WIN}{backslash_backup_path}"
        backslash_backup_root = os.path.dirname(backup_path).replace('/', '\\')
        tmp_backup_root = f"{TASK_TMP_FILES_DIR_WIN}{backslash_backup_root}"
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'copying backup to node host'
    duration_step_start = time.perf_counter()

    jsn = json.loads(remote_command_win(host, f'if not exist "{tmp_backup_path}" mkdir "{tmp_backup_path}"', True), strict=False)
    if jsn['stderr'] != '':
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        return

    is_error, result = copy_files_linux_win(backup_host, host, backup_path, tmp_backup_root, False)
    if is_error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,result,1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,result,1)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    
    step = 'restoring the database'
    duration_step_start = time.perf_counter()
    
    try:
        if compress:
            jsn = json.loads(remote_command_win(host, f'tar -xzkf "{tmp_backup_path}\\{backup_name}.tar.gz" -C "{backup_path}"', True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            is_error, result = restore_sqlserver_database(host, node, db_name, tmp_backup_path, backup_name, BackupType.FULL.value, BACKUP_CERTIFICATE['password'], restore_date=restore_date, norecovery=False)
            if is_error:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(result),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(result),1)
                return
            
        remote_command_win(host, f'rd /s /q "{tmp_backup_path}"', True)

        sql_conn = connect_to_sqlserver_instance(host['name'], 'master', node['database_user'], node['database_pwd'], node['port'])
        if isinstance(sql_conn, tuple):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(sql_conn[1]),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(sql_conn[1]),1)
            return
        
        with sql_conn:
            with sql_conn.cursor() as sql_cur:
                sql_cur.execute("SELECT SUM(mf.size * 8) AS SizeKB FROM sys.master_files mf JOIN sys.databases d ON mf.database_id = d.database_id WHERE mf.database_id > 4 GROUP BY mf.database_id")
                res = sql_cur.fetchone()
                node_size = res[0] if res and res[0] else 0.0
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
        
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                if result > 0.0:
                    cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='true', ind_deleted='false', size=%s WHERE id=%s""", (Severity.OK.value, Status.OK.value, result, db_id))
                    cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true', ind_deleted='false', size=%s WHERE id=%s""", [Severity.OK.value, Status.OK.value, node_size, node['id']])
                    conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def export_sqlserver_database_task(data, task_id): 
    """Export the database by DAC sqlpackage utility"""

    task = 'export sqlserver database task'
    
    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()
    
    try:
        host = json.loads(data['host_data'])
        node = json.loads(data['node_data'])
        backup_host = json.loads(data['backup_host'])
        db_data = json.loads(data['db_data'])
        db_name = db_data['name'] 
        backup = json.loads(data['backup_data'])
        backup_path = backup['ex_params'].get("backup_path")
        backup_name = backup['name']
        compress = backup['ex_params'].get("compress")
        include_data = data['include_data']
        backup_file = f'{backup_path}/{backup_name}.{"b" if include_data else "d"}acpac' 
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
	
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
	
    step = 'exporting the database'
    duration_step_start = time.perf_counter()
    
    try:
        jsn = json.loads(remote_command(backup_host, f"sudo mkdir -p '{backup_path}'", True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])

        sqlpackage_bin = find_sql_package(backup_host)
        commands = [
            f"sudo {sqlpackage_bin}/sqlpackage /Quiet /a:Ex{'port' if include_data else 'tract'} /TargetFile:'{backup_file}' /SourceConnectionString:'SERVER={host['name']},{node['port']};DATABASE={db_name};UID={node['database_user']};PWD={node['database_pwd']};TrustServerCertificate=yes' /p:VerifyExtraction=true {'/p:ExtractAllTableData=false' if not include_data else ''}"
        ]
        if compress:
            backup_tar = f'{backup_path}/{backup_name}.tar.gz'
            commands.append(f"sudo tar zcfC '{backup_tar}' '{backup_path}' '{os.path.basename(backup_file)}' --remove-files")
        
        is_error, result = share_database_access(host, node, db_name)
        if is_error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,result,1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,result,1)
            return
        
        for command in commands:
            jsn = json.loads(remote_command(backup_host, command, True), strict=False)
            if jsn['stderr'] != '' and "successfully exported database and saved it to file" not in jsn['stderr'].lower():
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return

            if jsn['stdout'] != '':
                print(jsn['stdout'])
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'getting the backup size'
    duration_step_start = time.perf_counter()

    try:
        jsn = json.loads(remote_command(backup_host, f"sudo du -b '{backup_file if not compress else backup_tar}' | awk '{{print $1}}'", True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        backup_size = float(jsn['stdout']) / 1024.0
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = jsonb_set(ex_params, '{{size}}', '{0}', true), status='true' WHERE id={1}""".format(backup_size,backup['id']))
                conn.commit()	
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def import_sqlserver_instance_task(data, task_id):
    """Import databases of instance by DAC sqlpackage utility"""
    
    task = 'import sqlserver instance task'
    
    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()
    
    try:
        host = json.loads(data['host_data'])
        node = json.loads(data['node_data']) 
        backup_host = json.loads(data['backup_host'])
        backup = json.loads(data['backup_data'])
        backup_path = backup['ex_params'].get("backup_path")
        compress = backup['ex_params'].get("compress")
        backup_name = backup['name']
        extension = '.dacpac'
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    step = 'importing node databases'
    duration_step_start = time.perf_counter()
    #this step will never be red,since import of just some databases is better than nothing
    last_known_error = None

    try:
        if compress:
            jsn = json.loads(remote_command(backup_host, f"sudo tar xzfC '{backup_path}/{backup_name}.tar.gz' '{backup_path}'", True), strict=False)
            if jsn['stderr'] != '':
                last_known_error = jsn['stderr']
                logging.warning(F"{task_id}: {task}: {step}: {db_name}: {jsn['stderr']}")

        jsn = json.loads(remote_command(backup_host, f"sudo ls -lt --file-type '{backup_path}' | grep '\{extension}$' | awk '{{print $9}}'", True), strict=False)
        export_filenames = [f for f in jsn['stdout'].splitlines() if f.split('.')[0].lower() not in SYSTEM_DATABASES]
        commands = []
        sqlpackage_bin = find_sql_package(backup_host)
        for filename in export_filenames:
            db_name = filename.split('.')[0]
            commands.append(f"sudo {sqlpackage_bin}/sqlpackage /Quiet /a:Publish /SourceFile:'{backup_path}/{filename}' /TargetConnectionString:'SERVER={host['name']},{node['port']};DATABASE={db_name};UID={node['database_user']}; PWD={node['database_pwd']};TrustServerCertificate=yes' /p:VerifyDeployment=true")
            is_error, result = ensure_database_exists(host, node, db_name)
            if is_error:
                last_known_error = jsn['stderr']
                logging.warning(F"{task_id}: {task}: {step}: {db_name}: {result}")
            
            is_error, result = share_database_access(host, node, db_name)
            if is_error:
                last_known_error = jsn['stderr']
                logging.warning(F"{task_id}: {task}: {step}: {db_name}: {result}")
        
        for command in commands:
            jsn = json.loads(remote_command(backup_host, command, True), strict=False)
            if jsn['stderr'] != '':
                last_known_error = jsn['stderr']
                logging.warning(F"{task_id}: {task}: {step}: {command}: {jsn['stderr']}")
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    if last_known_error:
        status = Severity.WARNING.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
    else:
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true' WHERE id=%s""", (Severity.OK.value, Status.OK.value, node['id'],))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def import_sqlserver_database_task(data, task_id):
    """import the database by DAC sqlpackage utility"""

    task = 'import sqlserver database task'
    
    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return
    
    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()
    
    try:
        host = json.loads(data['host_data'])
        backup_host = json.loads(data['backup_host'])
        node = json.loads(data['node_data'])
        db_data = json.loads(data['db_data'])
        db_name = db_data['name']
        backup = json.loads(data['backup_data'])
        backup_path = backup['ex_params'].get("backup_path")
        backup_name = backup['name']
        compress = backup['ex_params'].get("compress")
        include_data = backup['ex_params'].get("include_data")
        backup_file = f"{backup_path}/{backup_name}"
        extension = f"{'b' if include_data else 'd'}acpac"
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
	
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'importing the database'
    duration_step_start = time.perf_counter()

    try:
        is_error, result = drop_database(host, node, db_name, recreate=True)
        if is_error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,result,1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,result,1)
            return
        
        sqlpackage_bin = find_sql_package(backup_host)
        commands = [
            f"sudo {sqlpackage_bin}/sqlpackage /Quiet /a:{'Import' if include_data else 'Publish'} /SourceFile:'{backup_file}.{extension}' /TargetConnectionString:'SERVER={host['name']},{node['port']};DATABASE={db_name};UID={node['database_user']};PWD={node['database_pwd']};TrustServerCertificate=yes' {'/p:VerifyDeployment=true' if not include_data else ''}",
        ]
        if compress:
            commands.insert(0, f'tar xzfC "{backup_file}.tar.gz" "{backup_path}"')

        for command in commands:
            jsn = json.loads(remote_command(backup_host, command, True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
            
        if compress:
            #when clear temporary files - dont care of success
            remote_command(backup_host, f"sudo rm -f '{backup_file}.{extension}'", True)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'getting the database size'
    duration_step_start = time.perf_counter()
    
    try:
       sql_conn = connect_to_sqlserver_instance(host['name'], db_name, node['database_user'], node['database_pwd'], node['port'])
       if isinstance(sql_conn, tuple):
           status = Severity.NOT_OK.value
           duration_step_end = time.perf_counter() - duration_step_start
           logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(sql_conn[1]),1)
           step = 'null'
           status = Severity.NOT_OK.value
           duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
           logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(sql_conn[1]),1)
           return
       else:
          with sql_conn:
              with sql_conn.cursor() as sql_cur:
                    sql_cur.execute('''SELECT SUM(size * 8) / 1024  AS size_kb
                        FROM sys.master_files
                        WHERE database_id = DB_ID()
                        GROUP BY database_id''')
                    res = sql_cur.fetchone()
                    size = float(res[0]) if res and res[0] else 0.0
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_started='true', ind_deleted='false', size=%s WHERE id=%s""", (Severity.OK.value, Status.OK.value, size, db_data['id']))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def export_sqlserver_instance_task(data, task_id):
    """Export databases of instance by DAC sqlpackage utility"""
    
    task = 'export sqlserver instance task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()
    
    try:
        global SYSTEM_DATABASES
        host = json.loads(data['host_data'])
        node = json.loads(data['node_data']) 
        backup_host = json.loads(data['backup_host'])
        backup = json.loads(data['backup_data'])
        backup_path = backup['ex_params'].get("backup_path")
        compress = backup['ex_params'].get('compress')
        backup_name = backup['name']
        include_data = data['include_data']
        extension = '.dacpac'
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    step = 'exporting the databases'
    duration_step_start = time.perf_counter()

    commands = []
    try:
        sql_conn = connect_to_sqlserver_instance(host['name'], 'master', node['database_user'], node['database_pwd'], node['port'])
        if isinstance(sql_conn, tuple):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(sql_conn[1]),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(sql_conn[1]),1)
            return
        else:
            with sql_conn:
                with sql_conn.cursor() as sql_cur:
                    sql_cur.execute("SELECT name FROM master.sys.databases WHERE name NOT IN ({}) AND state = 0 AND is_in_standby = 0".format(','.join("?"*len(SYSTEM_DATABASES))), SYSTEM_DATABASES)
                    db_names = [name[0] for name in  sql_cur.fetchall()]
        
        jsn = json.loads(remote_command(backup_host, f"sudo mkdir -p '{backup_path}'", True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
        
        modified_date = get_date_modified_folder(backup_host, backup_path)
        if isinstance(modified_date, tuple):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(modified_date[1]),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(modified_date[1]),1)
            return
        
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = ex_params || '{"modified_date":\"%s\"}' WHERE id=%s""" % (modified_date, backup['id']))
                conn.commit()

        sqlpackage_bin = find_sql_package(backup_host) 
        for db_name in db_names:
            backup_file = f'{backup_path}/{db_name}{extension}'
            commands.append(f"sudo {sqlpackage_bin}/sqlpackage /Quiet /a:Extract /TargetFile:'{backup_file}' /SourceConnectionString:'SERVER={host['name']},{node['port']};DATABASE={db_name};UID={node['database_user']};PWD={node['database_pwd']};TrustServerCertificate=yes' /p:ExtractAllTableData={'true' if include_data else 'false'} /p:VerifyExtraction=true")
            is_error, result = share_database_access(host, node, db_name)
            #red, since unable to change db access mode may indicate corruption
            if is_error:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,result,1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,result,1)
                return

        for command in commands:
            jsn = json.loads(remote_command(backup_host, command, True), strict=False)
            #red, since if any error - entire database of node will be lost from export
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
        
        if compress:
            command = f"sudo tar zcfC '{backup_path}/{backup_name}.tar.gz' '{backup_path}' $(sudo ls -lt --file-type '{backup_path}' | grep '\{extension}$' | awk '{{print $9}}') --remove-files"
            jsn = json.loads(remote_command(backup_host, command, True), strict=False)
            if jsn['stderr'] != '':
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
        
        jsn = json.loads(remote_command(backup_host, f"sudo du -bs '{backup_path}' | awk '{{print $1}}'", True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        backup_size = float(jsn['stdout']) / 1024.0
        
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = jsonb_set(ex_params, '{{size}}', '{0}', true), status='true' WHERE id={1}""".format(backup_size,backup['id']))
                    cur.execute("""UPDATE "clusterCore_c123_inventory" SET ex_params = jsonb_set(ex_params, '{{duration}}', '{0}', true) WHERE id={1}""".format(duration_task_end,backup['id']))
                    conn.commit()
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def snapshot_sqlserver_database_task(data, task_id):
    
    task = 'snapshot sqlserver database task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()
    
    try:
        host = json.loads(data['host_data'])
        node = json.loads(data['node_data'])
        db = json.loads(data['db'])
        backup = json.loads(data['backup'])
        db_name = db['name']
        db_id = db['id']
        snapshot_id = backup['id']
        backup_path = backup["ex_params"].get("backup_path")
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'getting all data files of database'
    duration_step_start = time.perf_counter()
    
    sql_conn = connect_to_sqlserver_instance(host['name'], 'master', node['database_user'], node['database_pwd'], node['port'])
    if isinstance(sql_conn, tuple):
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(sql_conn[1]),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(sql_conn[1]),1)
        return
    
    try:
        sql_conn.autocommit = True
        sql_cur = sql_conn.cursor()
        sql_cur.execute(f"USE [{db_name}]")
        sql_cur.execute("SELECT name FROM sys.database_files")
        fetch_data = sql_cur.fetchall()
        file_names = [file_name[0] for file_name in fetch_data if file_name[0]]
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    step = 'creating snapshot of database'
    duration_step_start = time.perf_counter()

    sql_conn = connect_to_sqlserver_instance(host['name'], 'master', node['database_user'], node['database_pwd'], node['port'])
    if isinstance(sql_conn, tuple):
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(sql_conn[1]),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(sql_conn[1]),1)
        return
    
    try:
        sql_conn.autocommit = True
        sql_cur = sql_conn.cursor()
        all_files_commands = [f"NAME = '{string}', FILENAME = '{backup_path}\\{string}_' + CONVERT(VARCHAR(50), GETDATE(), 20) + '.ss'" for string, _ in file_names]
        create_snapshot_command = f"CREATE DATABASE [{db_name}_snapshot] ON ("
        create_snapshot_command += "),\n(".join(all_files_commands)
        create_snapshot_command += f") AS SNAPSHOT OF {db_name}"
        sql_cur.execute(create_snapshot_command)
        sql_cur.execute(f'USE {db_name}_snapshot')
        sql_cur.execute("SELECT SUM(size * 8) / 1024 AS size_kb FROM sys.master_files WHERE database_id = DB_ID() GROUP BY database_id")
        res = sql_cur.fetchone()
        size = float(res[0]) if res and res[0] else 0.0
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_inventory" SET type=%s, severity=%s, status=%s, size = %s, database_id =%s WHERE id =%s """, [Inventory.EXPORT.value, Severity.OK.value, Status.OK.value, size, db_id,snapshot_id])
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)
    
def replication_sqlserver_database_task(data, task_id):
    
    task = 'replication sqlserver database task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        repl_type = data["ex_params"].get("replication_type")
        distributor_host = json.loads(data['distributor_host_data'])
        distributor = json.loads(data['distributor_data']) 
        distribution_database_name = distributor['ex_params'].get("distribution_database_name")
        # common directory for all servers in cluster 
        repl_directory = f"{TASK_TMP_FILES_DIR_WIN}\\replDirectory"
        publisher_host = json.loads(data['publisher_host_data'])
        publisher = json.loads(data['publisher_data'])
        subscriber_host = json.loads(data['subscriber_host_data'])
        subscriber = json.loads(data['subscriber_data'])
        db = json.loads(data['db_data'])
        # name of replicated database on subscriber
        replication_name = json.loads(data["replication_name"])
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'configuring the repldata directory'
    duration_step_start = time.perf_counter()

    try:
        commands = [
            f'if not exist "{repl_directory}" mkdir "{repl_directory}"',
            f"net share repldirectory=drive:\"{repl_directory}\" /GRANT:{distributor['database_user']}, FULL /GRANT:{publisher['database_user']}, FULL /GRANT:{subscriber['database_user']}, READ"
        ]
        for command in commands:
            jsn = json.loads(remote_command_win(distributor_host, command, True), strict=False)
            if jsn['stderr'] != "":
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'creating the distributor'
    duration_step_start = time.perf_counter()

    sql_distributor_conn = connect_to_sqlserver_instance(distributor_host['name'], 'master', distributor['database_user'], distributor['database_pwd'], distributor['port'])
    if isinstance(sql_distributor_conn, tuple):
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(sql_distributor_conn[1]),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(sql_distributor_conn[1]),1)
        return
    
    try:
        sql_distributor_conn.autocommit = True
        sql_distributor_cur = sql_distributor_conn.cursor()
        sql_distributor_cur.execute('EXEC sp_get_distributor')
        is_installed = bool(sql_distributor_cur.fetchone[0])
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    if is_installed:
        try:
            # remove replication database first
            sql_publisher_cur.execute(f'USE [{db["name"]}]')
            sql_publisher_cur.execute(f"""EXEC sp_removedbreplication N'{db["name"]}'""")
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return
        
        try:
            # remove distributor
            sql_distributor_cur.execute(f'''USE master
                EXEC sp_dropdistpublisher {publisher["name"]}
                EXEC sp_dropdistributiondb {distribution_database_name}
                EXEC sp_dropdistributor
                ''')
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return
    
    try:
        # configure distributor
        sql_distributor_cur.execute(f'''
        DECLARE @distributor AS sysname
        DECLARE @distributorlogin AS sysname
        DECLARE @distributorpassword AS sysname
        
        SET @distributor = @@servername
        SET @distributorlogin = N'{distributor["database_user"]}'
        SET @distributorpassword = N'{distributor["database_pwd"]}'

        USE master
        EXEC sp_adddistributor @distributor = @distributor
        EXEC sp_adddistributiondb @database = N'distribution', @log_file_size = 2, @deletebatchsize_xact = 5000, @deletebatchsize_cmd = 2000, @security_mode = 0, @login = @distributorlogin, @password = @distributorpassword

        DECLARE @snapshotdirectory AS nvarchar(500)
        SET @snapshotdirectory = N'{repl_directory}'

        USE [distribution] 
        IF (NOT EXISTS (SELECT * FROM sysobjects WHERE name = 'UIProperties' AND type = 'U ')) 
        CREATE TABLE UIProperties(id int) 
        IF (EXISTS (SELECT * FROM ::fn_listextendedproperty('SnapshotFolder', 'user', 'dbo', 'table', 'UIProperties', null, null))) 
        EXEC sp_updateextendedproperty N'SnapshotFolder', @snapshotdirectory, 'user', dbo, 'table', 'UIProperties' 
        ELSE 
        EXEC sp_addextendedproperty N'SnapshotFolder', @snapshotdirectory, 'user', dbo, 'table', 'UIProperties'
        ''')
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'creating a publisher'
    duration_step_start = time.perf_counter()
    
    sql_publisher_conn = connect_to_sqlserver_instance(publisher_host['name'], 'master', publisher['database_user'], publisher['database_pwd'], publisher['port'])
    if isinstance(sql_distributor_conn, tuple):
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(sql_distributor_conn[1]),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(sql_distributor_conn[1]),1)
        return
    
    try:
        sql_publisher_conn.autocommit = True
        sql_publisher_cur = sql_publisher_conn.cursor()
        sql_publisher_cur.execute(f'''
        DECLARE @publisher AS sysname
        DECLARE @distributorlogin AS sysname
        DECLARE @distributorpassword AS sysname
        
        SET @publisher = @@Servername 
        SET @distributorlogin = N'{distributor['database_user']}'
        SET @distributorpassword = N'{distributor['database_pwd']}'
        
        EXEC sp_adddistpublisher @publisher = @publisher,
        @distribution_db = N'{distribution_database_name}',
        @security_mode = 0, 
        @login = @distributorlogin, 
        @password = @distributorpassword, 
        @working_directory = N'{repl_directory}', 
        @trusted = N'true', 
        @thirdparty_flag = 0, 
        @publisher_type = N'MSSQLSERVER'
        '''
        )
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'creating a database on subscriber'
    duration_step_start = time.perf_counter()

    sql_subscriber_conn = connect_to_sqlserver_instance(subscriber_host["name"], 'master', subscriber["database_user"], subscriber["database_pwd"], subscriber["port"])
    if isinstance(sql_subscriber_conn, tuple):
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(sql_subscriber_conn[1]),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(sql_subscriber_conn[1]),1)
        return
    
    try:
        sql_subscriber_conn.autocommit = True
        sql_subscriber_cur = sql_subscriber_conn.cursor()
        sql_subscriber_cur.execute(f"CREATE DATABASE [{db['name']}]")
        sql_subscriber_cur.execute("SELECT @@servername")
        subscriber_name = str(sql_subscriber_cur.fetchone()[0])
        
        sql_publisher_cur.execute("SELECT @@servername")
        publisher_name = str(sql_publisher_cur.fetchone()[0])
        
        sql_distributor_cur.execute("SELECT @@servername")
        distributor_name = str(sql_distributor_cur.fetchone()[0])
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'configuring publication job'
    duration_step_start = time.perf_counter()
    
    try:
        if repl_type == 'merge':
            add_publication_procedure = f"sp_add{repl_type}publication"
        else:
            add_publication_procedure = "sp_addpublication"
        
        publication_settings = f'''
        @publication = N'{db["name"]}_pub', 
        @description = N'{repl_type} publication of database '{db["name"]}' from Publisher',
        @retention = 0, 
        @allow_push = N'true',
        @allow_pull = N'true', 
        @repl_freq = N'snapshot', 
        @status = N'active', 
        @independent_agent = N'true'
        '''
        
        if repl_type == "merge":
            publication_settings += f''',
            @allow_subscription_copy = N'false',
            @add_to_active_directory = N'false',
            @dynamic_filters = N'false',
            @conflict_retention = 14,
            @keep_partition_changes = N'false',
            @allow_synctoalternate = N'false',
            @max_concurrent_merge = 0,
            @max_concurrent_dynamic_snapshots = 0,
            @use_partition_groups = N'false',
            @publication_compatibility_level = N'100RTM',
            @replicate_ddl = 1,
            @allow_subscriber_initiated_snapshot = N'false',
            @allow_web_synchronization = N'false',
            @allow_partition_realignment = N'true',
            @retention_period_unit = N'days',
            @conflict_logging = N'both',
            @automatic_reinitialization_policy = 0
        '''
        
        sql_publisher_cur.execute(f'''
        DECLARE @replicationdb AS sysname
        DECLARE @publisherlogin AS sysname
        DECLARE @publisherpassword AS sysname
        SET @replicationdb = N'{replication_name}'
        SET @publisherlogin = N'{publisher["database_user"]}'
        SET @publisherpassword = N'{publisher["database_pwd"]}'

        USE [{db['name']}]
        EXEC sp_replicationdboption @dbname = N'{db["name"]}', @optname = N'publish', @value = N'true'

        EXEC {add_publication_procedure}
        {publication_settings}

        EXEC sp_addpublication_snapshot @publication = N'{db["name"]}_pub',
        @frequency_type = 1,
        @frequency_interval = 1,
        @frequency_relative_interval = 1,
        @frequency_recurrence_factor = 0,
        @frequency_subday = 8,
        @frequency_subday_interval = 1,
        @active_start_time_of_day = 0,
        @active_end_time_of_day = 235959,
        @active_start_date = 0,
        @active_end_date = 0,
        @publisher_security_mode = 0,
        @publisher_login = @publisherlogin,
        @publisher_password = @publisherpassword
        '''
        )
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
   
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'creating articles for publication'
    duration_step_start = time.perf_counter()

    try:
        sql_publisher_cur.execute(create_new_subscription(db=db, subscriber_sql_name=subscriber_name, subscriber_user=subscriber["database_user"], subscriber_pwd=subscriber["database_pwd"], publication_name=f'{db["name"]}_pub', publication_type=repl_type, publication_operation="Push"))
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
   
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'finding out the job names'
    duration_step_start = time.perf_counter()

    try:
        sql_publisher_cur.execute('SELECT name, date_modified FROM msdb.dbo.sysjobs ORDER BY date_modified DESC')
        job_names = sql_publisher_cur.fetchall()
        publisher_job_name = next((i[0] for i in job_names if db['name'] in i[0] and publisher_name in i[0]), None)
        
        sql_distributor_cur.execute('SELECT name, date_modified FROM msdb.dbo.sysjobs ORDER BY date_modified DESC')
        job_names = sql_distributor_cur.fetchall()
        distributor_job_name = next((i[0] for i in job_names if db['name'] in i[0] and distributor_name in i[0]), None)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'generating the snapshot'
    duration_step_start = time.perf_counter()

    try:
        sql_publisher_cur.execute('USE msdb')
        sql_publisher_cur.execute(f"EXEC dbo.sp_start_job N'{publisher_job_name}'")

        sql_distributor_cur.execute('USE msdb')
        sql_distributor_cur.execute(f"EXEC dbo.sp_start_job N'{distributor_job_name}'")
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""INSERT INTO "clusterCore_c123_database" 
                        ("name", "type", created_date, last_update_date, size, ind_deployed, ind_started, ind_autostartup, ind_deleted, severity, sizing, database_user, database_pwd, created_by_id, last_update_by_id, node_id) 
                        VALUES (%s, %s, now(), now(), %s, 'false', 'true', 'false', 'false', %s, 'MEDIUM', %s, %s, %s, %s, %s)""", (replication_name, DBType.MSSQL.value, db["size"], Severity.OK.value, subscriber['database_user'], subscriber['database_pwd'], created_by, created_by, subscriber['id']))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)
    
def add_subscriber_to_sqlserver_publication_task(data, task_id):
    
    task = 'add subscriber to sqlserver publication task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()

    try:
        publisher_host = json.loads(data['publisher_host_data'])
        publisher = json.loads(data['publisher_data'])
        subscriber_host = json.loads(data['subscriber_host_data'])
        subscriber = json.loads(data['subscriber_data'])
        db = json.loads(data['db_data'])
        # name of replicated database on subscriber
        replication_name = json.loads(data["replication_name"])
        publication_name = json.loads(data["publication_name"])
        # pull or push
        publication_operation = json.loads(data["publication_operation"])
        #'transactional' or 'merge' #TODO add more types if exists
        publication_type = json.loads(data["publication_type"])
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'adding the new subscriber'
    duration_step_start = time.perf_counter()

    subscriber_conn = connect_to_sqlserver_instance(subscriber_host["name"], "master", subscriber["database_user"], subscriber["database_pwd"], subscriber["port"])
    if isinstance(subscriber_conn, tuple):
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(subscriber_conn[1]),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(subscriber_conn[1]),1)
        return
    
    try:
        subscriber_conn.autocommit = True
        subscriber_cur = subscriber_conn.cursor()
        subscriber_cur.execute("SELECT @@servername")
        subscriber_name = str(subscriber_cur.fetchone()[0])
        subscriber_cur.execute(f"IF NOT EXISTS (SELECT * FROM sys.databases WHERE name = '{replication_name}' BEGIN CREATE DATABASE [{replication_name}] END")
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    publisher_conn = connect_to_sqlserver_instance(publisher_host["name"], "master", publisher["database_user"], publisher["database_pwd"], publisher["port"])
    if isinstance(publisher_conn, tuple):
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(publisher_conn[1]),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(publisher_conn[1]),1)
        return
    
    try:
        publisher_conn.autocommit = True
        publisher_cur = publisher_conn.cursor()
        publisher_cur.execute(create_new_subscription(db, subscriber_name, replication_name, subscriber["database_user"], subscriber["database_pwd"], publication_name, publication_operation=publication_operation, publication_type=publication_type))
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
   
    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""INSERT INTO "clusterCore_c123_database" 
                        ("name", "type", created_date, last_update_date, size, ind_deployed, ind_started, ind_autostartup, ind_deleted, severity, sizing, database_user, database_pwd, created_by_id, last_update_by_id, node_id) 
                        VALUES (%s, %s, now(), now(), %s, 'false', 'true', 'false', 'false', %s, 'MEDIUM', %s, %s, %s, %s, %s)""", (replication_name, DBType.MSSQL.value, db["size"], Severity.OK.value, subscriber['database_user'], subscriber['database_pwd'], created_by, created_by, subscriber['id']))
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def restore_sqlserver_instance_task(data, task_id):
    """restore SQL Server instance from backups of all (non-system) databases"""
    
    task = 'restore sqlserver instance task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()
    
    try:
        global SYSTEM_DATABASES
        global BACKUP_CERTIFICATE
        host = json.loads(data['host_data'])
        node = json.loads(data['node_data'])
        backup = json.loads(data['backup_data'])
        backup_host = json.loads(data['backup_host'])
        compress = backup['ex_params'].get("compress")
        backup_path = backup['ex_params'].get("backup_path")
        restore_date = data['date']
        restore_option = data['restore_option'].strip().lower()
        backslash_backup_path = backup_path.replace('/', '\\')
        tmp_backup_path = f"{TASK_TMP_FILES_DIR_WIN}{backslash_backup_path}"
        backslash_backup_root = os.path.dirname(backup_path).replace('/', '\\')
        tmp_backup_root = f"{TASK_TMP_FILES_DIR_WIN}{backslash_backup_root}"
        backup_name = backup['name']
        # tar is available by default in windows 10/11 only
        full_archive = f"{tmp_backup_path}\\{backup_name}@full.tar.gz"
        log_archive = f"{tmp_backup_path}\\{backup_name}@log.tar.gz"
        
        if RestoreOptions.THIS.value in restore_option:
            # like "latest" in postgresql - ignore point in time
            restore_date = ''
        else:
            """
            timezone obtaining are temporary disabled, 
            since real datetime of our windows machines does not match its timezone
            """ 
            # timezone_cmd = '[Windows.Globalization.Calendar,Windows.Globalization,ContentType=WindowsRuntime]::New().GetTimeZone()'
            # jsn = json.loads(remote_command_win(host, timezone_cmd, True, 'ps'), strict=False)
            # if int(jsn['excode']) != 0:
            #     status = Severity.NOT_OK.value
            #     duration_step_end = time.perf_counter() - duration_step_start
            #     logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            #     step = 'null'
            #     status = Severity.NOT_OK.value
            #     duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            #     logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            #     return

            # restore_date = datetime.strptime(restore_date, "%Y-%m-%d %H:%M:%S").astimezone(pytz.timezone(jsn['stdout'])).strftime('%Y-%m-%d %H:%M:%S') if RestoreOptions.LATEST.value not in restore_option.lower() else ''
            
            # latest = current time
            if RestoreOptions.LATEST.value in restore_option:
                # safe, since restoring to any time after backup completed is no harm
                restore_date = datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S')
            # any else = user-defined time
            else:
                restore_date = datetime.strptime(restore_date, "%Y-%m-%d %H:%M:%S").astimezone(pytz.UTC).strftime('%Y-%m-%d %H:%M:%S')    
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'copying backups to node host'
    duration_step_start = time.perf_counter()
    
    try:
        jsn = json.loads(remote_command_win(host, f'if not exist "{tmp_backup_path}" mkdir "{tmp_backup_path}"', True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        # is_error, result = copy_files_linux_win(backup_host, host, backup_path, tmp_backup_root, False)
        # if is_error:
        #     status = Severity.NOT_OK.value
        #     duration_step_end = time.perf_counter() - duration_step_start
        #     logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,result,1)
        #     step = 'null'
        #     status = Severity.NOT_OK.value
        #     duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        #     logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,result,1)
        #     return
        
        last_known_err = ''
        if compress:
            jsn = json.loads(remote_command_win(host, f'''tar -xzkf "{full_archive}" -C "{tmp_backup_path}"''', True, 'ps'), strict=False)
            if int(jsn['excode']) != 0:
                #throw red step, since if log are not preceeding by full - task are bugging
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
            
            if RestoreOptions.THIS.value not in restore_option:
                jsn = json.loads(remote_command_win(host, f'''if (Test-Path -Path "{log_archive}") {{ tar -xzkf "{log_archive}" -C "{tmp_backup_path}" }};''', True, 'ps'), strict=False)
                if int(jsn['excode']) != 0:
                    # trigger yellow step, log, proceed
                    last_known_err = jsn['stderr']
                    logging.error(f"{task}: {task_id}: {step}: {last_known_err}")

        jsn = json.loads(remote_command_win(host,f"""Get-ChildItem -Path "{tmp_backup_path}" | Where-Object {{$_.Name -match '.bak|.trn'}} | Select-Object -ExpandProperty Name""", True, 'ps'), strict=False)
        if int(jsn['excode']) != 0:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
        
        backup_filenames = {i.strip() for i in jsn['stdout'].splitlines() if re.search(r'\S', i)}
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return

    status = Severity.OK.value if not last_known_err else 'yellow'
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by) if status == 'green' else logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,last_known_err['stdout'],last_known_err['stderr'],last_known_err['excode'])

    step = 'restoring instance databases'
    duration_step_start = time.perf_counter()
    
    try:
        #Partial backups should have?Also could be applied native mssql COMPRESSION...
        """
        Backups types restore sequence memo:
        1.full or file/filegroup (with norecovery, 
        if there are unapplied log or differential or file/filegroup backups exists for same db)
        2.differential (with norecovery, 
        if there are unapplied log backups exists for same db)
        3.log (with recovery)
        4.others
        """
        db_sizes = {}
        section_groups = {
            BackupType.FULL.value: [],
            BackupType.DIFFERENTIAL.value: [],
            BackupType.LOG.value: [],
            "others": []
        }
        for i, parts in zip(backup_filenames, (i.split('@') for i in backup_filenames)):
            name = parts[0]
            backup_type = parts[1].split('.')[0].lower()
            backup_name = i

            section = {"name": name, 'backup_type': backup_type, 'backup_name': backup_name}

            if BackupType.FULL.value in backup_type:
                section_groups[BackupType.FULL.value].append(section)
            elif BackupType.DIFFERENTIAL.value in backup_type:
                section_groups[BackupType.DIFFERENTIAL.value].append(section)
            elif BackupType.LOG.value in backup_type:
                section_groups[BackupType.LOG.value].append(section)
            else:
                section_groups["others"].append(section)
        
        for group_name in section_groups:
            section_groups[group_name] = sorted(section_groups[group_name], key=lambda i: i['name'])
        
        marker = '{}@'
        # stop immediately with red step if at least one database failed to restore
        for db in section_groups[BackupType.FULL.value]:
            norecovery = RestoreOptions.THIS.value not in restore_option and (any(i['backup_name'].startswith(marker.format(db['name'])) for i in section_groups[BackupType.DIFFERENTIAL.value]) or any(i['backup_name'].startswith(marker.format(db['name'])) for i in section_groups[BackupType.LOG.value]))
            is_error, result = restore_sqlserver_database(host, node, db['name'], tmp_backup_path, db['backup_name'], db['backup_type'], BACKUP_CERTIFICATE['password'], restore_date=restore_date, norecovery=norecovery)
            if is_error:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(result),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(result),1)
                return
            
            db_sizes[db['name']] = result
        
        if RestoreOptions.THIS.value not in restore_option:
            for db in section_groups[BackupType.DIFFERENTIAL.value]:
                norecovery = any(i['backup_name'].startswith(marker.format(db['name'])) for i in section_groups[BackupType.LOG.value])
                is_error, result = restore_sqlserver_database(host, node, db['name'], tmp_backup_path, db['backup_name'], db['backup_type'], BACKUP_CERTIFICATE['password'], restore_date=restore_date, norecovery=norecovery)
                if is_error:
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(result),1)
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(result),1)
                    return
                
                db_sizes[db_name] = result

            for db in section_groups[BackupType.LOG.value]:
                is_error, result = restore_sqlserver_database(host, node, db['name'], tmp_backup_path, db['backup_name'], db['backup_type'], BACKUP_CERTIFICATE['password'], restore_date=restore_date, norecovery=False)
                if is_error:
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(result),1)
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(result),1)
                    return
                
                db_sizes[db['name']] = result
            
        for db in section_groups['others']:
            is_error, result = restore_sqlserver_database(host, node, db['name'], tmp_backup_path, db['backup_name'], db['backup_type'], BACKUP_CERTIFICATE['password'], restore_date=restore_date, norecovery=False)
            if is_error:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(result),1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(result),1)
                return
            
            db_sizes[db['name']] = result

        sql_conn = connect_to_sqlserver_instance(host['name'], 'master', node['database_user'], node['database_pwd'], node['port'])
        if isinstance(sql_conn, tuple):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(sql_conn[1]),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(sql_conn[1]),1)
            return
        
        with sql_conn:
            with sql_conn.cursor() as sql_cur:
                sql_cur.execute("SELECT SUM(mf.size * 8) AS SizeKB FROM sys.master_files mf JOIN sys.databases d ON mf.database_id = d.database_id WHERE mf.database_id > 4 GROUP BY mf.database_id")
                res = sql_cur.fetchone()
                node_size = res[0] if res and res[0] else 0.0
    
    #    remote_command_win(host, f'rd /s /q "{tmp_backup_path}"', True)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                for db_name, db_size in db_sizes.items():
                    cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, size=%s, ind_deleted='false' WHERE node_id =%s AND name=%s""", [Severity.OK.value, Status.OK.value, db_size, node['id'], db_name])
                    conn.commit()
                
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true', ind_deleted='false', size=%s WHERE id=%s""", [Severity.OK.value, Status.OK.value, node_size, node['id']])
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
   
    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)

def mirroring_sqlserver_database_task(data, task_id):
    """mirroring 'primary' db to 'partner' db"""

    task = 'mirroring sqlserver database task'

    try:
        data = json.loads(data)
        created_by = int(data['created_by'])
    except Exception as error:
        logging.error(f"{task}: {str(error)}")
        return

    task_log_id = uuid.uuid4().hex
    step = 'null'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    duration_task_start = datetime.now()

    step = 'initializing task'
    status = Severity.INFO.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)

    step = 'checking parameters'
    duration_step_start = time.perf_counter()
    
    try:
        host_source = json.loads(data['host_source'])
        node_source = json.loads(data['node_source'])
        host_target = json.loads(data['host_target'])
        node_target = json.loads(data['node_target'])
        db_exists = data['db_exists']
        db_source = json.loads(data['db_source'])
        db_source_name = db_source['name']
        db_target = json.loads(data['db_target'])
        db_source_id = db_source['id']
        db_target_name = db_target['name']
        # default port = 5022
        source_endpoint_port = data['source_endpoint_port']
        target_endpoint_port = data['target_endpoint_port']
        temp_backup_full_path = f'{TASK_TMP_FILES_DIR_WIN}\\mirroring\\{node_source["name"].replace(".", "_").replace(":", "_")}\\full'
        backup_full_file = f'{temp_backup_full_path}\\{db_source_name}.bak'
        is_async = data['is_async']
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    #MUST go first, since on incompatible instances there are no point to do other preparations
    step = 'checking instances compatibility'
    duration_step_start = time.perf_counter()
    
    sql_source_conn = connect_to_sqlserver_instance(host_source['name'], db_source_name, node_source['database_user'], node_source['database_pwd'], node_source['port'])
    if isinstance(sql_source_conn, tuple):
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(sql_source_conn[1]),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(sql_source_conn[1]),1)
        return
    
    try:
        sql_source_conn.autocommit = True
        sql_source_cur = sql_source_conn.cursor()
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    if node_source['name'] == node_target['name'] and node_source['port'] == node_target['port']:
        sql_target_conn = sql_source_conn
        sql_target_cur = sql_source_cur
    else:
        sql_target_conn = connect_to_sqlserver_instance(host_target['name'], 'master', node_target['database_user'], node_target['database_pwd'], node_target['port'])
        if isinstance(sql_source_conn, tuple):
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(sql_source_conn[1]),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(sql_source_conn[1]),1)
            return

        try:
            sql_target_conn.autocommit = True
            sql_target_cur = sql_target_conn.cursor()
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return
    
    is_error, source_major_version = get_major_version(sql_source_conn)
    if is_error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(source_major_version),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(source_major_version),1)
        return

    is_error, destination_major_version = get_major_version(sql_target_conn)
    if is_error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(destination_major_version),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(destination_major_version),1)
        return
    
    if int(destination_major_version) != int(source_major_version):
        error = f"Mirroring could not be performed.\nDestination instance version {destination_major_version} are not equal to the source instance version {source_major_version}."
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,1)
        return
    
    is_error, source_edition_id = get_edition_id(sql_source_conn)
    if is_error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(source_edition_id),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(source_edition_id),1)
        return

    is_error, destination_edition_id = get_edition_id(sql_target_conn)
    if is_error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(destination_edition_id),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(destination_edition_id),1)
        return
    
    if source_edition_id != destination_edition_id:
        error = f"Mirroring could not be performed.\nDestination instance edition ID {destination_edition_id} are not equal to the source instance edition ID {source_edition_id}"
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,1)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    step = 'checking prerequisites for mirroring'
    duration_step_start = time.perf_counter()

    # checking if users has sysadmin role and dbs are not in mirroring already
    sysadmin_query = "SELECT IS_SRVROLEMEMBER('sysadmin')"
    for db_user in [node_source['database_user'], node_target['database_user']]:
        try:
            if db_user == node_source['database_user']:
                sql_cur = sql_source_cur
                node_name = node_source["name"]
            else:
                sql_cur = sql_target_cur
                node_name = node_target["name"]
                
            sql_cur.execute(sysadmin_query)
            sysadmin_result = sql_cur.fetchone()

            if sysadmin_result:
                sysadmin = int(sysadmin_result[0])

                if sysadmin != 1:
                    error = f"Mirroring could not be performed.\nDatabase user {db_user} of instance {node_name} do not have sysadmin role."
                    status = Severity.NOT_OK.value
                    duration_step_end = time.perf_counter() - duration_step_start
                    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,1)
                    step = 'null'
                    status = Severity.NOT_OK.value
                    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,1)
                    return
            else:
                error = f'Mirroring could not be performed.\nDatabase user {db_user} is not found on instance {node_name}.'
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,1)
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,1)
                return
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return

    mirror_query = "SELECT mirroring_guid FROM sys.database_mirroring WHERE database_id = DB_ID(?)"
    recovery_mode_query = "SELECT compatibility_level, recovery_model_desc FROM sys.databases WHERE name = ?"
    for db_name in db_source_name, db_target_name:
        params = (db_name,)

        if db_name == db_source_name:
            node_name = node_source["name"]
            sql_cur = sql_source_cur
        else:
            node_name = node_source["name"]
            sql_cur = sql_target_cur

        try:
            sql_cur.execute(mirror_query, params)
            mirror_result = sql_cur.fetchone()

            sql_cur.execute(recovery_mode_query, params)
            recovery_mode_result = sql_cur.fetchone()
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return
        
        #if no rows - database are (likely) does not exist on instance
        if not mirror_result:
            error = f"Mirroring could not be performed.\nDatabase {db_name} is not found on instance {db_name}."
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,1)
            return
        
        # Reverse logic here - expect to return None
        if mirror_result[0]:
            error = f"Database {db_name} of instance {node_name} is already in mirroring."
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,1)
            return
        
        if recovery_mode_result and len(recovery_mode_result) > 1:
            if db_name == db_source_name:
                compatibility_level_source = str(recovery_mode_result[0])
                compatibility_level_target = compatibility_level_source
            else:
                compatibility_level_target = str(recovery_mode_result[0])

            recovery_model = str(recovery_mode_result[1])
        else:
            recovery_model = ValidationResult.ERROR.value

        if RecoveryModel.FULL.value not in recovery_model.lower():
            error = f"Mirroring could not be performed.\nDatabase {db_name} of instance {node_name} recovery mode is not {RecoveryModel.FULL.name}."
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,1)
            return

    if compatibility_level_source != compatibility_level_target:
        error = f"Mirroring could not be performed.\nCompatibility level of database {db_source_name} {compatibility_level_source} are not equal to compatibility level of database {db_target_name} {compatibility_level_target}."
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,error,1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,error,1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'creating endpoints'
    duration_step_start = time.perf_counter()

    # Create a database mirroring endpoint CREATE ENDPOINT Mirroring
    # query = 'CREATE ENDPOINT {}_{} STATE=STARTED AS TCP (LISTENER_PORT={}) FOR DATABASE_MIRRORING (ROLE=PARTNER)'
    query = 'CREATE ENDPOINT {}_{} STATE=STARTED AS TCP (LISTENER_PORT={}) FOR DATABASE_MIRRORING (ROLE = ALL)'
    run_timestamp = datetime.now().strftime('%d%m%Y%H%M%S')
    try:
        #for testing purpose only
        sql_source_cur.execute('''DECLARE @DropEndpointSql NVARCHAR(MAX) = ''
SELECT @DropEndpointSql += 'DROP ENDPOINT ' + QUOTENAME(name) + ';'
FROM sys.database_mirroring_endpoints
WHERE type_desc = 'DATABASE_MIRRORING'

IF @DropEndpointSql <> ''
    EXEC (@DropEndpointSql)''')
        sql_source_cur.execute(query.format(db_source_name, run_timestamp, source_endpoint_port))
        #for testing purpose only
        sql_target_cur.execute('''DECLARE @DropEndpointSql NVARCHAR(MAX) = ''
SELECT @DropEndpointSql += 'DROP ENDPOINT ' + QUOTENAME(name) + ';'
FROM sys.database_mirroring_endpoints
WHERE type_desc = 'DATABASE_MIRRORING'

IF @DropEndpointSql <> ''
    EXEC (@DropEndpointSql)''')
        sql_target_cur.execute(query.format(db_target_name, run_timestamp, target_endpoint_port))
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'backing up the source'
    duration_step_start = time.perf_counter()

    backup_full_query = f"BACKUP DATABASE ? TO DISK=? WITH FORMAT, INIT"
    source_full_params = [db_source_name, backup_full_file]
    
    try:
        jsn = json.loads(remote_command_win(host_source, f'if not exist "{temp_backup_full_path}" mkdir "{temp_backup_full_path}"', True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    try:
        sql_source_cur.execute(backup_full_query, source_full_params)
        wait_busy_cursor(cur)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'copying source backups to the target host'
    duration_step_start = time.perf_counter()

    client_target = create_winrm_connection(host_target['name'], host_target['username'], host_target['password'], host_target['ssh_port'])
    if isinstance(client_target, tuple):
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(client_target[1]),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(client_target[1]),1)
        return

    try:
        jsn = json.loads(remote_command_win(host_target, f'if not exist "{temp_backup_full_path}" mkdir "{temp_backup_full_path}"', True), strict=False)
        if jsn['stderr'] != '':
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
            return

        commands = [
            f'''
            $passwd = convertto-securestring -AsPlainText -Force -String "{host_target["password"]}"
            $cred = new-object -typename System.Management.Automation.PSCredential -argumentlist "{host_target["username"]}@{host_target["ex_params"].get("domain")}",$passwd
            $spn = New-PSSessionOption -IncludePortInSPN
            $s = New-PSSession -ComputerName {host_target["name"]} -Port {host_target["ssh_port"]} -Credential $cred -SessionOption $spn
            Invoke-Command -ScriptBlock {{Copy-Item -Path '{backup_full_file}' -Destination '{backup_full_file}' -ToSession $s}}'''
        ]
        for command in commands:
            jsn = json.loads(remote_command_win(host_source, command, True, "ps"))
            if jsn['excode'] != 0:
                status = Severity.NOT_OK.value
                duration_step_end = time.perf_counter() - duration_step_start
                logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                step = 'null'
                status = Severity.NOT_OK.value
                duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
                logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,jsn['stdout'],jsn['stderr'],jsn['excode'])
                return
            remote_command_win(host_source, f'rd /s /q "{temp_backup_full_path}"', True)
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),9)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),9)
        return
    
    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    step = 'restoring the destination'
    duration_step_start = time.perf_counter()
    
    state_query = "ALTER DATABASE [{}] SET {} WITH ROLLBACK IMMEDIATE"
    recovery_full_query = "RESTORE DATABASE ? FROM DISK = ? WITH REPLACE, NORECOVERY"
    target_full_params = [db_target_name, backup_full_file]
    partner_query = "ALTER DATABASE [{}] SET PARTNER = 'TCP://{}:{}'"
    # Set the partner instance on the mirror server
    try:
        # On potentially existing db - always grant exclusive access,THEN put offline
        sql_target_cur.execute(f"ALTER DATABASE [{db_target_name}] SET single_user WITH ROLLBACK IMMEDIATE")
        sql_target_cur.execute(state_query.format(db_target_name, DatabaseState.OFFLINE.name))
        sql_target_cur.execute(recovery_full_query, target_full_params)
        wait_busy_cursor(cur)
        # need manually take in online state for this restore, since restore WITH NORECOVERY
        sql_target_cur.execute(state_query.format(db_target_name, DatabaseState.ONLINE.name))
        sql_target_cur.execute(partner_query.format(db_target_name, host_source['name'], source_endpoint_port))
        sql_source_cur.execute(partner_query.format(db_source_name, host_target['name'], target_endpoint_port))
        sql_target_cur.execute("SELECT ROUND(SUM(size * 8 / 1024.0), 2) AS [Size_KB] FROM sys.databases AS db JOIN sys.master_files AS mf ON db.database_id = mf.database_id WHERE db.name = ?", db_target['name'])
        res = sql_target_cur.fetchone()
        db_size = float(res[0]) if res and res[0] else 0.0
        remote_command_win(host_target, f'rd /s /q "{temp_backup_full_path}"', True)
        sql_target_cur.execute("SELECT SUM(mf.size * 8) AS SizeKB FROM sys.master_files mf JOIN sys.databases d ON mf.database_id = d.database_id WHERE mf.database_id > 4 GROUP BY mf.database_id")
        res = sql_target_cur.fetchone()
        node_size = res[0] if res and res[0] else 0.0
    except Exception as error:
        status = Severity.NOT_OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return

    status = Severity.OK.value
    duration_step_end = time.perf_counter() - duration_step_start
    logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)

    if is_async:
        # By default, SAFETY is ON, which means the mode is synchronous
        step = 'configuring asynchronous database mirroring'
        duration_step_start = time.perf_counter()

        query = f"ALTER DATABASE [{db_source_name}] SET PARTNER SAFETY OFF"

        try:
            sql_source_cur.execute(query)
        except Exception as error:
            status = Severity.NOT_OK.value
            duration_step_end = time.perf_counter() - duration_step_start
            logger(task_id,task_log_id,task,step,status,duration_step_end,created_by,None,str(error),1)
            step = 'null'
            status = Severity.NOT_OK.value
            duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
            logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
            return
        
        status = Severity.OK.value
        duration_step_end = time.perf_counter() - duration_step_start
        logger(task_id,task_log_id,task,step,status,duration_step_end,created_by)
    
    try:
        with get_postgres_connection() as conn:
            with conn.cursor() as cur:
                cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_deleted='false' WHERE id=%s""", (Severity.OK.value, Status.OK.value, db_source_id,))
                cur.execute("""SELECT id FROM "clusterCore_c123_database" WHERE node_id=%s AND name =%s""", (node_target['id'], db_target_name))
                if cur.rowcount < 1:
                    cur.execute("""INSERT INTO "clusterCore_c123_database" 
                            ("name", "type", created_date, last_update_date, size, ind_deployed, ind_started, ind_autostartup, ind_deleted, severity, sizing, database_user, database_pwd, created_by_id, last_update_by_id, node_id) 
                            VALUES (%s, %s, now(), now(), %s, 'false', 'true', 'false', 'false', %s, 'MEDIUM', %s, %s, %s, %s, %s)""", (db_target_name, DBType.MSSQL.value, db_size, Severity.OK.value, node_target["database_user"], node_target["database_pwd"], created_by, created_by, node_target['id']))
                else:
                    db_target_id = cur.fetchone()[0]
                    cur.execute("""UPDATE "clusterCore_c123_database" SET severity=%s, status=%s, ind_deleted='false', size=%s WHERE id=%s""", (Severity.OK.value, Status.OK.value, db_size, db_target_id,))
                
                cur.execute("""UPDATE "clusterCore_c123_node" SET severity=%s, status=%s, ind_started='true', ind_deleted='false', size=%s WHERE id=%s""", [Severity.OK.value, Status.OK.value, node_size, node_target['id']])
                conn.commit()
    except Exception as error:
        step = 'null'
        status = Severity.NOT_OK.value
        duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
        logger(task_id,task_log_id,task,step,status,duration_task_end,created_by,None,str(error),1)
        return
   
    step = 'task finished'
    status = Severity.OK.value
    duration = 0
    logger(task_id,task_log_id,task,step,status,duration,created_by)
    step = 'null'
    status = Severity.OK.value
    duration_task_end = round((datetime.now() - duration_task_start).total_seconds())
    logger(task_id,task_log_id,task,step,status,duration_task_end,created_by)